#!/usr/bin/env python
# coding=utf-8
#pylint: disable=too-many-public-methods,too-many-locals,too-many-arguments,superfluous-parens,import-error,invalid-name,wildcard-import,unused-wildcard-import,missing-docstring
from ctypes import c_uint8, c_uint32
from tool.device.nvme import NVMe
from tool.device.nvme import struct as STRUCT
from tool.device.nvme.struct import FeatureID as Fid
from tool.device.nvme import buf


class Feature(object):
    def __init__(self, dev, fid, dw11=0, nsid=0):
        self.nsid = nsid
        self.fid = fid
        self.dw11 = dw11
        self.dev = NVMe() if dev is None else dev

        self.dword0 = buf.Malloc(length=1, types=c_uint32)
        self.buffer = buf.Malloc(length=4096 * 2, types=c_uint8)
        self.get_cap()
        self.base_values = {}

    def get_cap(self):
        """get capability"""
        cdword0_struct = self.dev.get_features(self.fid, 3, self.dw11)[1]
        cdword0 = cdword0_struct.DWORD
        self.saveable = cdword0 & 0x1
        self.changeable = (cdword0 & 0b100) >> 2
        self.ns_spec = (cdword0 & 0b10) >> 1

    def get_expect_value(self, sel, set_attr):
        """get expect value"""
        if 'save' in set_attr:
            set_sv = set_attr['save']
        else:
            set_sv = 0
        if sel == 0:
            if self.changeable:
                return set_attr
            return self.base_values[0]
        elif sel == 1:
            return self.base_values[1]
        elif sel == 2:
            if (set_sv is 0) or (not self.saveable):
                return self.base_values[2]
            return set_attr
        else:
            print("Not supported SEL in compare mode")
        return False

    def printf(self):
        pass

    def check_attr(self, set_attr, sel):
        """check attribute"""
        expect = self.get_expect_value(sel, set_attr)
        if not expect:
            return expect
        print('actual value:')
        self.printf()
        print('expect value:\n{}'.format(expect))
        return self.compare_field(expect)

    def compare_field(self, expect):
        pass


class Arbitration(Feature):
    """Arbitration base on Feature"""
    def __init__(self, dev=None, nsid=0):
        super(Arbitration, self).__init__(dev, Fid.Arbitration, nsid=nsid)
        self.cfg = self.dword0.convert(STRUCT.Arbitration)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"hpw": dw0.HPW,
                     "mpw": dw0.MPW,
                     "lpw": dw0.LPW,
                     "ab": dw0.AB}
            self.base_values[i] = value

    def printf(self):
        """print Arbitration"""
        print("---------> get feature Arbitration")
        print(" Namespace:                                 0x{:x}".format(self.nsid))
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" High Priority Weight:                      0x{:x}".format(self.cfg.HPW))
        print(" Medium Priority Weight:                    0x{:x}".format(self.cfg.MPW))
        print(" Low Priority Weight:                       0x{:x}".format(self.cfg.LPW))
        print(" Arbitration Burst:                         0x{:x}".format(self.cfg.AB))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_arbitration(set_attr['save'], set_attr['ab'], set_attr['lpw'],
                                                set_attr['mpw'], set_attr['hpw'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.HPW == expect['hpw'] and self.cfg.MPW == expect['mpw'] and \
                self.cfg.LPW == expect['lpw'] and self.cfg.AB == expect['ab']:
            return True
        return False


class PowerManagement(Feature):
    """Power Management (Feature Identifier 02h)"""

    def __init__(self, dev=None):
        super(PowerManagement, self).__init__(dev, Fid.PowerManagement)
        self.cfg = self.dword0.convert(STRUCT.PowerManagement)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"wh": dw0.WH,
                     "ps": dw0.PS}
            self.base_values[i] = value

    def printf(self):
        """print value"""
        print("---------> Get Feature - Power Management")
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Workload Hint:                             0x{:x}".format(self.cfg.WH))
        print(" Power State:                               0x{:x}".format(self.cfg.PS))

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.WH == expect['wh'] and self.cfg.PS == expect['ps']:
            return True
        return False

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_power_management(set_attr['save'],
                                                     set_attr['ps'], set_attr['wh'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret


class LBARangeType(Feature):
    """LBA Range Type (Feature Identifier 03h), (Optional)"""

    def __init__(self, dev=None, prpo1=0, prpo2=0, index=0):
        super(LBARangeType, self).__init__(dev, Fid.LBARangeType, prpo1, prpo2)
        self.cfg = self.dword0.convert(STRUCT.LBARangeType)
        self.entry = self.buffer.convert(STRUCT.LBARangeTypeEntry, index=index)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            value = {"TYPE": self.entry.TYPE, "ATTR": self.entry.ATTR, "SLBA": self.entry.SLBA,
                     "NLB": self.entry.NLB, "GUID": self.entry.GUID.LOW}
            self.base_values[i] = value
            print(self.base_values)

    def printf(self):
        """print value"""
        print("---------> get feature LBA Range Type")
        print(" Saveable:                    0x{:x}".format(self.saveable))
        print(" Changeable:                  0x{:x}".format(self.changeable))
        print(" Namespace specific:          0x{:x}".format(self.ns_spec))
        print(" Type:                        0x{:x}".format(self.entry.TYPE))
        print(" Attributes:                  0x{:x}".format(self.entry.ATTR))
        print(" Starting LBA:                0x{:x}".format(self.entry.SLBA))
        print(" Number of Logical Blocks:    0x{:x}".format(self.entry.NLB))
        print(" Unique Identifier:           0x{!r}".format(self.entry.GUID.LOW))

    def test_values(self, set_attr):
        """test values"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        lrt = [{'TYPE': 3, 'ATTR': 2, 'SLBA': 0x1234, 'NLB': 0x321, 'GUID': 0x2}]
        ret = self.dev.set_features_lba_range_type(lrt, set_attr['save'])
        for i in range(3):
            self.entry = self.dev.get_features(self.fid, i, self.dw11)[2][0]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.entry.TYPE == expect['TYPE'] and self.entry.ATTR == expect['ATTR'] and \
                self.entry.SLBA == expect['SLBA'] and self.entry.NLB == expect['NLB'] and \
                self.entry.GUID.LOW == expect['GUID']:
            return True
        return False


class TemperatureThreshold(Feature):
    """TemperatureThreshold base on Feature"""

    def __init__(self, dev=None):
        super(TemperatureThreshold, self).__init__(dev, Fid.TemperatureThreshold, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.TemperatureThreshold)
        self.set_attr = None
        self.get_current_value()

    def printf(self):
        """print value"""
        print("---------> get feature Temperature Threshold")
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print("Threshold Type Select:                      0x{:x}".format(self.cfg.THSEL))
        print("Threshold Temperature Select:               0x{:x}".format(self.cfg.TMPSEL))
        print("Temperature Threshold:                      0x{:x}".format(self.cfg.TMPTH))

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"thsel": dw0.THSEL,
                     "tmpsel": dw0.TMPSEL,
                     "tmpth": dw0.TMPTH}
            self.base_values[i] = value

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_temperature_threshold(set_attr['save'], set_attr['tmpth'],
                                                          set_attr['tmpsel'], set_attr['thsel'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.THSEL == expect['thsel'] and self.cfg.TMPSEL == expect['tmpsel'] and \
                self.cfg.TMPTH == expect['tmpth']:
            return True
        return False


class ErrorRecovery(Feature):
    """ErrorRecovery base on Feature"""

    def __init__(self, dev=None):
        super(ErrorRecovery, self).__init__(dev, Fid.ErrorRecovery, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.ErrorRecovery)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"tler": dw0.TLER,
                     "dulbe": dw0.DULBE}
            self.base_values[i] = value
            print(self.base_values)

    def printf(self):
        """print value"""
        print(" Saveable:               0x{:x}".format(self.saveable))
        print(" Changeable:             0x{:x}".format(self.changeable))
        print(" Namespace specific:     0x{:x}".format(self.ns_spec))
        print(" Deallocated or Unwritten Logical Block Error Enable: 0x{:x}".format(self.cfg.DULBE))
        print(" Time Limited Error Recovery:  0x{:x}".format(self.cfg.TLER))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_error_recovery(set_attr['save'],
                                                   set_attr['tler'], set_attr['dulbe'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.DULBE == expect['dulbe'] and self.cfg.TLER == expect['tler']:
            return True
        return False


class VolatileWriteCache(Feature):
    """VolatileWriteCache base on Feature"""

    def __init__(self, dev=None):
        super(VolatileWriteCache, self).__init__(dev, Fid.VolatileWriteCache, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.VolatileWriteCache)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            valuse = {"wce": dw0.WCE}
            self.base_values[i] = valuse

    def printf(self):
        """print value"""
        print("---------> get feature Volatile WriteCache")
        print(" Saveable:             0x{:x}".format(self.saveable))
        print(" Changeable:           0x{:x}".format(self.changeable))
        print(" Namespace specific:   0x{:x}".format(self.ns_spec))
        print(" Volatile Write Cache Enable:     0x{:x}".format(self.cfg.WCE))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_volatile_write_cache(set_attr['save'], set_attr['wce'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.WCE == expect['wce']:
            return True
        return False


class InterruptCoalescing(Feature):
    """InterruptCoalescing"""

    def __init__(self, dev=None):
        super(InterruptCoalescing, self).__init__(dev, Fid.InterruptCoalescing, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.InterruptCoalescing)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            values = {"time": dw0.TIME, "thr": dw0.THR}
            self.base_values[i] = values
            print(self.base_values)

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Aggregation Time:                          0x{:x}".format(self.cfg.TIME))
        print(" Aggregation Threshold:                     0x{:x}".format(self.cfg.THR))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_interrupt_coalescing(set_attr['save'],
                                                         set_attr['thr'], set_attr['time'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.THR == expect['thr'] and self.cfg.TIME == expect['time']:
            return True
        return False


class InterruptVectorConfiguration(Feature):
    """InterruptVectorConfiguration"""

    def __init__(self, dev=None):
        super(InterruptVectorConfiguration, self).__init__(dev, Fid.InterruptVectorConfiguration,
                                                           0, 0)
        self.cfg = self.dword0.convert(STRUCT.InterruptVectorConfiguration)
        self.get_current_value()

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Coalescing Disable:                        0x{:x}".format(self.cfg.CD))
        print(" Interrupt Vector:                          0x{:x}".format(self.cfg.IV))

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"cd": dw0.CD, "iv": dw0.IV}
            self.base_values[i] = value
            print(self.base_values)

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_interrupt_vector_configuration(set_attr['save'],
                                                                   set_attr['iv'], set_attr['cd'])
        for i in range(3):
            iv = [0x4, 0x9, 0x10]
            self.dw11 = iv[i]
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.CD == expect['cd'] and self.cfg.IV == expect['iv']:
            return True
        return False


class WriteAtomicity(Feature):
    """WriteAtomicity"""

    def __init__(self, dev=None):
        super(WriteAtomicity, self).__init__(dev, Fid.WriteAtomicityNormal, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.WriteAtomicityNormal)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"dn": dw0.DN}
            self.base_values[i] = value
            print(self.base_values)

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Disable Normal:                            0x{:x}".format(self.cfg.DN))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_write_atomicity_normal(set_attr['save'], set_attr['dn'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.DN == expect['dn']:
            return True
        return False


class AsynchronousEventConfiguration(Feature):
    """AsynchronousEventConfiguration"""

    def __init__(self, dev=None):
        super(AsynchronousEventConfiguration, self).__init__(dev,\
                                        Fid.AsynchronousEventConfiguration, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.AsynchronousEventConfiguration)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"fan": dw0.FAN,
                     "nan": dw0.NAN,
                     "shcw": dw0.SHCW,
                     "tln": dw0.TLN}
            self.base_values[i] = value

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Firmware Activation Notices:               0x{:x}".format(self.cfg.FAN))
        print(" Namespace Attribute Notices:               0x{:x}".format(self.cfg.NAN))
        print(" SMART / Health Critical Warnings:          0x{:x}".format(self.cfg.SHCW))

    def test_values(self, set_attr):
        """test_value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_asynchronous_event_configuration(set_attr['save'], \
                set_attr['shcw'], set_attr['nan'], set_attr['fan'], set_attr['tln'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.FAN == expect['fan'] and self.cfg.NAN == expect['nan'] and \
                self.cfg.SHCW == expect['shcw'] and self.cfg.TLN == expect['tln']:
            return True
        return False


class AutonomousPowerStateTransition(Feature):
    """Autonomous Power State Transition"""

    def __init__(self, dev=None):
        super(AutonomousPowerStateTransition, self).__init__(dev, Fid.AutonomousPowerStateTransition, 0, 0)
        self.entry = self.buffer.convert(STRUCT.AutonomousPowerStateTransitionEntry)
        self.cfg = self.dword0.convert(STRUCT.AutonomousPowerStateTransition)
        self.get_current_value()

    def printf(self):
        """print value"""
        print(" Saveable:                                   0x{:x}".format(self.saveable))
        print(" Changeable:                                 0x{:x}".format(self.changeable))
        print(" Namespace specific:                         0x{:x}".format(self.ns_spec))
        print(" Autonomous Power State Transition Enable:   0x{:x}".format(self.cfg.APSTE))
        print(" Idle Time Prior to Transition:              0x{:x}".format(self.entry.ITPT))
        print(" Idle Transition Power State:                0x{:x}".format(self.entry.ITPS))

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            valuse = {"apste": dw0.APSTE}
            self.base_values[i] = valuse

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.APSTE == expect['apste']:
            return True
        return False

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_autonomous_power_state_transition(set_attr['save'], set_attr['apste'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret


class SoftwareProgressMarker(Feature):
    """SoftwareProgressMarker base on Feature"""

    def __init__(self, dev=None):
        super(SoftwareProgressMarker, self).__init__(dev, Fid.SoftwareProgressMarker, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.SoftwareProgressMarker)
        self.get_current_value()

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Pre-boot Software Load Count:              0x{:x}".format(self.cfg.PBSLC))

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"pbslc": dw0.PBSLC}
            self.base_values[i] = value
            print(self.base_values[i])

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_software_progress_marker(set_attr['save'], set_attr['pbslc'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.PBSLC == expect['pbslc']:
            return True
        return False


class HostIdentifier(Feature):
    """Host Identifier ONLY 8 bytes data structure"""

    def __init__(self, dev=None):
        super(HostIdentifier, self).__init__(dev, Fid.HostIdentifier, 0, 0)
        self.entry = self.buffer.convert(STRUCT.HostIdentifierEntry)
        self.cfg = self.dword0.convert(STRUCT.HostIdentifier)
        self.get_current_value()

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Enable Host Identifier:                    0x{:x}".format(self.cfg.EXHID))
        print(" Host Identifier Low 64:                    0x{:x}".format(self.entry.HOSTIDL))
        print(" Host Identifier High 64:                   0x{:x}".format(self.entry.HOSTIDH))

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            value = {"exhid": dw0.EXHID,
                     "hostidl": self.entry.HOSTIDL,
                     "hostidh": self.entry.HOSTIDH, }
            self.base_values[i] = value
            print(self.base_values[i], "WWWW")

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_host_identifier(set_attr['exhid'], set_attr['hostidl'],
                                                    set_attr['hostidh'], set_attr['save'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.entry.HOSTIDL == expect['hostidl'] and self.entry.HOSTIDH == expect['hostidh']:
            return True
        return False


class Timestamp(Feature):
    """ReservationPersistence"""

    def __init__(self, dev=None):
        super(Timestamp, self).__init__(dev, Fid.Timestamp, 0, 0)
        self.entry = self.buffer.convert(STRUCT.Timestamp)
        self.cfg = self.dword0.convert(STRUCT.Timestamp)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            value = {"timestamp": self.entry.TIMESTAMP,
                     "synch": self.entry.SYNCH,
                     "to": self.entry.TO}
            self.base_values[i] = value
            print(self.base_values[i])

    def printf(self):
        """print value"""
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Timestamp:                                 0x{:x}".format(self.entry.TIMESTAMP))
        print(" Synch:                                     0x{:x}".format(self.entry.SYNCH))
        print(" Timestamp Origin:                          0x{:x}".format(self.entry.TO))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_timestamp(set_attr['timestamp'], set_attr['save'])
        for i in range(3):
            self.dword0 = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def get_expect_value(self, sel, set_attr):
        """get expect value"""
        if sel in (0, 1, 2):
            value = self.base_values[0]
        else:
            print("Not supported SEL in compare mode")
            value = False
        return value

    def compare_field(self, expect):
        """compare_field"""
        if self.entry.TIMESTAMP == expect['timestamp'] and self.entry.SYNCH == expect['synch'] and \
                self.entry.TO == expect['to']:
            return True
        return False


class HostControlledThermalManagement(Feature):
    """KeepAliveTimer"""

    def __init__(self, dev=None):
        super(HostControlledThermalManagement, self).__init__(dev, \
                            Fid.HostControlledThermalManagement, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.HostControlledThermalManagement)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            print("-----------sel : %d" % i)
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            valuse = {"tmt1": dw0.TMT1,
                      "tmt2": dw0.TMT2}
            self.base_values[i] = valuse

    def printf(self):
        """print value"""
        print("---------> get feature Number of Queues")
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Thermal Management Temperature 1:          0x{:x}".format(self.cfg.TMT1))
        print(" Thermal Management Temperature 2:          0x{:x}".format(self.cfg.TMT2))

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.TMT1 == expect['tmt1'] and self.cfg.TMT2 == expect['tmt2']:
            return True
        return False

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_host_controlled_thermal_management(set_attr['save'], \
                                set_attr['tmt2'], set_attr['tmt1'])
        for i in range(3):
            self.dword0 = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret


class NonOperationalPowerStateConfig(Feature):
    """NonOperationalPowerStateConfig"""

    def __init__(self, dev=None):
        super(NonOperationalPowerStateConfig, self).__init__(dev, Fid.NonOperationalPowerStateConfig, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.NonOperationalPowerStateConfig)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(4):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            valuse = {"NOPPME": dw0.NOPPME}
            self.base_values[i] = valuse

    def printf(self):
        """print value"""
        print("---------> get feature Number of Queues")
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Non-Operational Power State Permissive Mode Enable: 0x{:x}".format(self.cfg.NOPPME))

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.NOPPME == expect['NOPPME']:
            return True
        return False

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_nonoperational_power_state_config(set_attr['save'], set_attr['noppme'])
        for i in range(3):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret


class HostMemoryBuffer(Feature):
    """HostMemoryBuffer"""
    def __init__(self, dev=None):
        super(HostMemoryBuffer, self).__init__(dev, Fid.HostMemoryBuffer, 0, 0)
        self.cfg = self.dword0.convert(STRUCT.HostMemoryBuffer)
        self.get_current_value()

    def get_current_value(self):
        """get current value"""
        for i in range(2):
            _, dw0, _ = self.dev.get_features(self.fid, i, self.dw11)
            valuse = {"EHM": dw0.EHM,
                      "MR":  dw0.MR}
            self.base_values[i] = valuse

    def printf(self):
        """print value"""
        print("---------> get feature Number of Queues")
        print(" Saveable:                                  0x{:x}".format(self.saveable))
        print(" Changeable:                                0x{:x}".format(self.changeable))
        print(" Namespace specific:                        0x{:x}".format(self.ns_spec))
        print(" Enable Host Memory: 0x{:x}".format(self.cfg.EHM))
        print(" Memory Return: 0x{:x}".format(self.cfg.MR))

    def test_values(self, set_attr):
        """test value"""
        return self.set_and_get_all_sel(set_attr)

    def set_and_get_all_sel(self, set_attr):
        """set_and_get_all_sel"""
        ret = self.dev.set_features_host_memory_buffer(set_attr['save'], \
            set_attr['ehm'], set_attr['mr'], set_attr['esize'], set_attr['hdmlec'])
        for i in range(2):
            self.cfg = self.dev.get_features(self.fid, i, self.dw11)[1]
            ret = self.check_attr(set_attr, i)
            if ret is False:
                print("***********************Test failed: select is {}".format(i))
            else:
                print("***********************Test Passed: select is {}".format(i))
        return ret

    def compare_field(self, expect):
        """compare_field"""
        if self.cfg.EHM == expect['EHM'] and self.cfg.MR == expect['MR']:
            return True
        return False




from test_framework.dut.agent import Agent
from test_framework.dut.slot import Slot
from test_framework.database.dut_database import DutDatabase
from utils import log


class Dut(object):

    def __init__(self):
        self.agent = Agent()
        self.slot = Slot(self.agent)

    def refresh(self):
        log.INFO("Begin to refresh DUT")
        self.agent.refresh()
        self.slot.refresh()

    # def refresh_slots(self):
    #     slots = self.get_agent_slots()
    #     for slot in slots:
    #         slot_obj = Slot(slot[0], slot[1], slot[2])
    #         slot_obj.refresh()


if __name__ == '__main__':
    ag = Dut()
    ag.refresh()



#!/usr/bin/env python
# pylint: disable=wildcard-import,unused-wildcard-import
"""
Created on 2020/06/03

@author: yyang
"""

from .log_page import *
from .features import *
from .identify import *
from .command import *
from .memory import *



#!/usr/bin/env python
# pylint: disable=superfluous-parens,missing-docstring,too-few-public-methods
"""
Created on Apr 19, 2019

@author: yyang
"""
from ctypes import Structure, c_uint8, c_uint16, c_uint32, c_uint64, Union

from tool.device.nvme.ctype import Uint128


class OPCODE(object):
    DELETE_IO_SQ = 0x00
    CREATE_IO_SQ = 0x01
    GET_LOG_PAGE = 0x02
    DELETE_IO_CQ = 0x04
    CREATE_IO_CQ = 0x05
    IDENTIFY = 0x06
    ABORT = 0x08
    SET_FEATURES = 0x09
    GET_FEATURES = 0x0A
    ASYNC_EVENT_REQUEST = 0x0C
    NAMESPACE_MGT = 0x0D
    FIRMWARE_COMMIT = 0x10
    FIRMWARE_DOWNLOAD = 0x11
    DEVICE_SELF_TEST = 0x14
    NAMESPACE_ATTACH = 0x15
    DIRECTIVE_SEND = 0x19
    DIRECTIVE_RECEIVE = 0x1A
    VIRT_MANAGEMENT = 0x1C
    NVME_MI_SEND = 0x1D
    NVME_MI_RECEIVE = 0x1E
    FORMAT_NVM = 0x80
    SECURITY_SEND = 0x81
    SECURITY_RECV = 0x82
    SANITIZE = 0X84

    FLUSH = 0x00
    WRITE = 0x01
    READ = 0x02
    WRITE_UNCORRECTABLE = 0x04
    COMPARE = 0x05
    WRITE_ZEROES = 0x08
    DATASET_MGT = 0x09
    RESERVATION_REGISTER = 0x0D
    RESERVATION_REPORT = 0x0E
    RESERVATION_ACQUIRE = 0x11
    RESERVATION_RELEASE = 0x15

    CNEX_WRITE_MEM = 0x99
    CNEX_READ_MEM = 0x9A
    CNEX_CLEAR_GBB = 0xC3
    CNEX_VU_COMMAND = 0xC4

    AGING_MF = 0xEC

    KNUCKLE_VU_SUM = 0xEC
    KNUCKLE_VU_FW = 0xED
    KNUCKLE_VU_GET = 0xEE
    KNUCKLE_VU_NON = 0xEC
    KNUCKLE_VU_WR = 0xED
    KNUCKLE_VU_RD = 0xEE
    KNUCKLE_VU_RATE_LI = 0xEF

    KNUCKLE_VU_SUB_LVM = 0x3
    KNUCKLE_VU_SUB_QUEUE_STATE = 0x3

    ADMIN = {
        0x0: 'Delete IO SQ',
        0x1: 'Create IO SQ',
        0x2: 'Get Log Page',
        0x4: 'Delete IO CQ',
        0x5: 'Create IO CQ',
        0x6: 'Identify',
        0x8: 'Abort',
        0x9: 'Set Features',
        0xa: 'Get Features',
        0xc: 'Asynchronous Event Request',
        0xd: 'Namespace Management',
        0x10: 'Firmware_Commit',
        0x11: 'Firmware Image_Download',
        0x14: 'Device Self-test',
        0x15: 'Namespace Attachment',
        0x18: 'Keep Alive',
        0x19: 'Directive Send',
        0x1a: 'Directive Receive',
        0x1c: 'Virtualization Management',
        0x1d: 'NVMe-MI Send',
        0x1e: 'NVMe-MI Receive',
        0x7c: 'Doorbell Buffer Config',
        0x80: 'Format NVM',
        0x81: 'Security Send',
        0x82: 'Security Receive',
        0x84: 'Sanitize',
    }

    IO = {
        0x0: 'Flush',
        0x1: 'Write',
        0x2: 'Read',
        0x4: 'Write Uncorrectable',
        0x5: 'Compare',
        0x8: 'Write Zeroes',
        0x9: 'Dataset Management',
        0xd: 'Reservation Register',
        0xe: 'Reservation Report',
        0x11: 'Reservation Acquire',
        0x15: 'Reservation Release',
        0x99: 'CNEX Write Memory',
        0x9a: 'CNEX Read Memory'
    }


class DebugOPCODE(object):
    ASYNC_TRACE = 0x00
    GET_CQE = 0x01
    GET_SQE = 0x02


class LVMOPCODE(object):
    START_TRACK_IO = 0x1
    START_TRACK_MEM = 0x2
    STOP_TRACK = 0x3
    STOP_MWR = 0x4


class LVMTenantID(object):
    PHR = 0x0
    LVM_IO = 0x1
    LVM_MWR = 0x2
    VF1_IO = 0x10
    VF1_MWR = 0x11
    VF2_IO = 0x20
    VF2_MWR = 0x21
    CIAI = 0x3
    INVALID = 0x3f


class PNVMeDebug(Structure):
    _pack_ = 1
    _fields_ = [
        ('OPCODE', c_uint32),
        ('RSV', c_uint32),
        ('DATA', c_uint64),
    ]


class PNVMeIOCmd(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OPCODE', c_uint8),
            ('FLAGS', c_uint8),
            ('CID', c_uint16),
            ('NSID', c_uint32),
            ('CDW2', c_uint32),
            ('CDW3', c_uint32),
            ('META', c_uint64),
            ('PRP1', c_uint64),
            ('PRP2', c_uint64),
            ('SLBA', c_uint64),
            ('NLB', c_uint16),
            ('CTRL', c_uint16),
            ('DSMGMT', c_uint8),
            ('QID', c_uint8),
            ('DIR', c_uint16),
            ('REFTAG', c_uint32),
            ('APPTAG', c_uint16),
            ('APPMASK', c_uint16),
        ]

    class CDWS(Structure):
        _pack_ = 1
        _fields_ = [
            ('CDW', c_uint32 * 16),
        ]

    _anonymous_ = ('CDWS', 'BITS')
    _fields_ = [
        ('CDWS', CDWS),
        ('BITS', BITS),
    ]


class NVMeUserIO(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OPCODE', c_uint8),
            ('FLAGS', c_uint8),
            ('CONTROL', c_uint16),
            ('NBLOCKS', c_uint16),
            ('RSV', c_uint16),
            ('META', c_uint64),
            ('DATA', c_uint64),
            ('SLBA', c_uint64),
            ('DSMGMT', c_uint32),
            ('REFTAG', c_uint32),
            ('APPTAG', c_uint16),
            ('APPMASK', c_uint16),
        ]

    class CDWS(Structure):
        _pack_ = 1
        _fields_ = [
            ('CDW', c_uint32 * 12),
        ]

    _anonymous_ = ('CDWS', 'BITS')
    _fields_ = [
        ('CDWS', CDWS),
        ('BITS', BITS),
    ]


class NVMeSGLIO(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OPCODE', c_uint8),
            ('FLAGS', c_uint8),
            ('CID', c_uint16),
            ('NSID', c_uint32),
            ('CDW2', c_uint32*2),
            ('META', c_uint64),
            ('PRP1', c_uint64),
            ('PRP2', c_uint64),
            ('SLBA', c_uint64),
            ('NLB', c_uint16),
            ('CTRL', c_uint16),
            ('DSMGMT', c_uint8),
            ('QID', c_uint8),
            ('DIRECTIVE_SPEC', c_uint16),
            ('REFTAG', c_uint32),
            ('APPTAG', c_uint16),
            ('APPMASK', c_uint16),
            ('DATA_OFF', c_uint32),
            ('META_OFF', c_uint32),
            ('DATA_DES_NUM', c_uint64),
            ('META_DES_NUM', c_uint64),
            ('SGL_DATA_PTR', c_uint64),
            ('SGL_META_PTR', c_uint64),
        ]

    class CDWS(Structure):
        _pack_ = 1
        _fields_ = [
            ('CDW', c_uint32 * 26),
        ]

    _anonymous_ = ('CDWS', 'BITS')
    _fields_ = [
        ('CDWS', CDWS),
        ('BITS', BITS),
    ]


class NVMePassthruCmd(Union):
    """
    Take Care: different with SQE!
    cdw8 is metadata_len and cdw9 is data_len
    Add cdw16: timeout
    Add cdw17: dword0 of CQE
    """

    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OPCODE', c_uint8),
            ('FLAGS', c_uint8),
            ('RSV', c_uint16),
            ('NSID', c_uint32),
            ('CDW2', c_uint32),
            ('CDW3', c_uint32),
            ('META', c_uint64),
            ('DATA', c_uint64),
            ('META_LEN', c_uint32),
            ('DATA_LEN', c_uint32),
            ('CDW10', c_uint32),
            ('CDW11', c_uint32),
            ('CDW12', c_uint32),
            ('CDW13', c_uint32),
            ('CDW14', c_uint32),
            ('CDW15', c_uint32),
            ('TIMEOUT', c_uint32),
            ('RESULT', c_uint32),
        ]

    class CDWS(Structure):
        _pack_ = 1
        _fields_ = [
            ('CDW', c_uint32 * 18),
        ]

    _anonymous_ = ('CDWS', 'BITS')
    _fields_ = [
        ('CDWS', CDWS),
        ('BITS', BITS)
    ]


class RWDword12(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NLB', c_uint16),  # Number Of Logical Blocks (NLB)
            ('RSV0', c_uint16, 4),
            ('DTYPE', c_uint16, 4),  # Directive Type (DTYPE)
            ('RSV1', c_uint16, 1),
            ('DEAC', c_uint16, 1),  # Deallocate (DEAC)
            ('PRINFO', c_uint16, 4),  # Protection Information (PRINFO)
            ('FUA', c_uint16, 1),  # Force Unit Access (FUA)
            ('LR', c_uint16, 1),  # Limited Retry (LR)
        ]

    class WORD(Structure):
        _pack_ = 1
        _fields_ = [
            ('NBLOCKS', c_uint16),
            ('CONTROL', c_uint16),
        ]

    _anonymous_ = ('BITS', 'WORD')
    _fields_ = [
        ('DWORD', c_uint32),
        ('WORD', WORD),
        ('BITS', BITS),
    ]


class RWDword13(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('DSM', c_uint8),  # Dataset Management (DSM)
            ('RSV', c_uint8),
            ('DSPEC', c_uint16),  # Directive Specific (DSPEC)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class ReservationAcquire(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('RACQA', c_uint32, 3),  # Reservation Acquire Action (RACQA)
            ('IEKEY', c_uint32, 1),  # Ignore Existing Key (IEKEY)
            ('RSV0', c_uint32, 4),
            ('RTYPE', c_uint32, 8),  # Reservation Type (RTYPE)
            ('RSV1', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class ReservationAcquireDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('CRKEY', c_uint64),  # Current Reservation Key (CRKEY)
        ('PRKEY', c_uint64),  # Preempt Reservation Key (PRKEY)
    ]


class ReservationRegister(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('RREGA', c_uint32, 3),  # Reservation Register Action (RREGA)
            ('IEKEY', c_uint32, 1),  # Ignore Existing Key (IEKEY)
            ('RSV0', c_uint32, 26),
            ('CPTPL', c_uint32, 2),  # Change Persist Through Power Loss State (CPTPL)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class ReservationRegisterDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('CRKEY', c_uint64),  # Current Reservation Key (CRKEY)
        ('NRKEY', c_uint64),  # New Reservation Key (NRKEY)
    ]


class ReservationRelease(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('RRELA', c_uint32, 3),  # Reservation Release Action (RRELA)
            ('IEKEY', c_uint32, 1),  # Ignore Existing Key (IEKEY)
            ('RSV0', c_uint32, 4),
            ('RTYPE', c_uint32, 8),  # Reservation Type (RTYPE)
            ('RSV1', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class ReservationReleaseDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('CRKEY', c_uint64),  # Current Reservation Key (CRKEY)
    ]


class RegisteredControllerDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('CNTLID', c_uint16),  # Controller ID (CNTLID)
        ('RCSTS', c_uint8),  # Reservation Status (RCSTS)
        ('RSV0', c_uint8 * 5),
        ('HOSTID', c_uint64),  # Host Identifier (HOSTID)
        ('RKEY', c_uint64),  # Reservation Key (RKEY)
    ]


class ReservationStatusDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('GEN', c_uint32),  # Generation (GEN)
        ('RTYPE', c_uint8),  # Reservation Type (RTYPE)
        ('REGCTL', c_uint16),  # Number of Registered Controllers (REGCTL)
        ('RSV0', c_uint16),
        ('PTPLS', c_uint8),  # Persist Through Power Loss State (PTPLS)
        ('RSV1', c_uint8 * 14),
        ('RCDS', RegisteredControllerDataStructure * 16),  # Registered Controller Data Structure
    ]


class RegisteredControllerExtendedDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('CNTLID', c_uint16),  # Controller ID (CNTLID)
        ('RCSTS', c_uint8),  # Reservation Status (RCSTS)
        ('RSV0', c_uint8 * 5),
        ('RKEY', c_uint64),  # Reservation Key (RKEY)
        ('HOSTID', Uint128),  # Host Identifier (HOSTID)
        ('RSV1', c_uint8 * 32),
    ]


class ReservationStatusExtendedDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('GEN', c_uint32),  # Generation (GEN)
        ('RTYPE', c_uint8),  # Reservation Type (RTYPE)
        ('REGCTL', c_uint16),  # Number of Registered Controllers (REGCTL)
        ('RSV0', c_uint16),
        ('PTPLS', c_uint8),  # Persist Through Power Loss State (PTPLS)
        ('RSV1', c_uint8 * 54),  # Reservation Type (RTYPE)
        ('RCEDS', RegisteredControllerExtendedDataStructure * 16),  # Registered Controller Extended Data Structure
    ]


class CreateIOCQDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('QID', c_uint32, 16),  # Queue Identifier (QID)
            ('QSIZE', c_uint32, 16),  # Queue Size (QSIZE)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class CreateIOCQDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('PC', c_uint32, 1),  # Physically Contiguous (PC)
            ('IEN', c_uint32, 1),  # Interrupts Enabled (IEN)
            ('RSV', c_uint32, 14),
            ('IV', c_uint32, 16),  # Interrupt Vector (IV)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class CreateIOSQDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('QUD', c_uint32, 16),  # Queue Identifier (QID)
            ('QSIZE', c_uint32, 16),  # Queue Size (QSIZE)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class CreateIOSQDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('PC', c_uint32, 1),  # Physically Contiguous (PC)
            ('QPRIO', c_uint32, 2),  # Queue Priority (QPRIO)
            ('RSV', c_uint32, 13),
            ('CQID', c_uint32, 16),  # Completion Queue Identifier (CQID)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DeleteIOCQDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('QID', c_uint32, 16),  # Queue Identifier (QID)
            ('RSV', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DeleteIOSQDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('QID', c_uint32, 16),  # Queue Identifier (QID)
            ('RSV', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DeviceSelfTest(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('STC', c_uint32, 4),  # Self-test Code (STC)
            ('RSV', c_uint32, 28),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DirectiveDW11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('DOPER', c_uint32, 8),  # Directive Operation (DOPER)
            ('DTYPE', c_uint32, 8),  # Directive Type (DTYPE)
            ('DSPEC', c_uint32, 16),  # Directive Specific (DSPEC)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class NVMeMIDW10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NMSP0', c_uint32, 8),  # NVMe-MI Specific 0 (NMSP0)
            ('RSV', c_uint32, 24),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class FWCommitDW10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('FS', c_uint32, 3),  # Firmware Slot (FS)
            ('CA', c_uint32, 3),  # Commit Action (CA)
            ('RSV', c_uint32, 25),
            ('BPID', c_uint32, 1),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class FormatDword10(Union):
    """
    The Format NVM command uses the Command Dword 10 field.
    All other command specific fields are reserved.
    """

    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('LBAF', c_uint32, 4),  # LBA Format (LBAF)
            ('MSET', c_uint32, 1),  # Metadata Settings (MSET)
            ('PIF', c_uint32, 3),  # Protection Information (PIF)
            ('PIL', c_uint32, 1),  # Protection Information Location (PIL)
            ('SES', c_uint32, 3),  # Secure Erase Settings (SES)
            ('RSV', c_uint32, 20),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class AbortDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('SQID', c_uint32, 16),
            ('CID', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class IdentifyDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('CNS', c_uint32, 8),
            ('RSV', c_uint32, 8),
            ('CNTID', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class GetFeaturesDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('FID', c_uint32, 8),
            ('SEL', c_uint32, 3),
            ('RSV', c_uint32, 21),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class SetFeaturesDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('FID', c_uint32, 8),
            ('RSV', c_uint32, 23),
            ('SV', c_uint32, 1),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class AttachNSDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('SEL', c_uint32, 4),
            ('RSV', c_uint32, 28),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class SanitizeDW10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('SANACT', c_uint32, 3),  # Sanitize Action
            ('AUSE', c_uint32, 1),  # Allow Unrestricted Sanitize Exit
            ('OWPASS', c_uint32, 4),  # Overwrite Pass Count
            ('OIPBP', c_uint32, 1),  # Overwrite Invert Pattern Between Passes
            ('NDAS', c_uint32, 1),  # No Deallocate After Sanitize
            ('RSV', c_uint32, 22),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class SanitizeDW11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OVRPAT', c_uint32, 32),  # Overwrite Pattern
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class CompareDword15(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('ELBAT', c_uint32, 16),
            ('ELBATM', c_uint32, 16),

        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DsmDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NR', c_uint32, 8),  # Number of Ranges (NR)
            ('RSV', c_uint32, 24)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class DsmDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('IDR', c_uint32, 1),
            ('IDW', c_uint32, 1),
            ('AD', c_uint32, 1),
            ('RSV', c_uint32, 29)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class NamespaceManagementDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('SEL', c_uint32, 4),
            ('RSV', c_uint32, 28)
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class SecurityReceiveDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NSSF', c_uint32, 8),
            ('SPSP0', c_uint32, 8),
            ('SPSP1', c_uint32, 8),
            ('SECP', c_uint32, 8),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class SecuritySendDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NSSF', c_uint32, 8),
            ('SPSP0', c_uint32, 8),
            ('SPSP1', c_uint32, 8),
            ('SECP', c_uint32, 8),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class GetLogPagDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('LID', c_uint32, 8),
            ('LSP', c_uint32, 4),  # Log Specific Field (LSP)
            ('RSV0', c_uint32, 3),
            ('RAE', c_uint32, 1),  # Retain Asynchronous Event (RAE)
            ('NUMDL', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class GetLogPagDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NUMDU', c_uint32, 16),
            ('RSV', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class GetLogPagDword12(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('LPOL', c_uint32, 32),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class GetLogPagDword13(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('LPOU', c_uint32, 32),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class VirtualizationManagementDword10(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('ACT', c_uint32, 4),
            ('RSV0', c_uint32, 4),
            ('RT', c_uint32, 3),
            ('RSV1', c_uint32, 5),
            ('CNTLID', c_uint32, 16),

        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class RateLimitsDword0(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('OPC', c_uint32, 8),
            ('FUSE', c_uint32, 2),
            ('RSV0', c_uint32, 4),
            ('PSDT', c_uint32, 2),
            ('CID', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class RateLimitsDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('IOR_OVERFLOW_CTR', c_uint32, 16),
            ('IOW_OVERFLOW_CTR', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class VirtualizationManagementDword11(Union):
    class BITS(Structure):
        _pack_ = 1
        _fields_ = [
            ('NR', c_uint32, 16),
            ('RSV0', c_uint32, 16),
        ]

    _anonymous_ = ('BITS',)
    _fields_ = [
        ('DWORD', c_uint32),
        ('BITS', BITS),
    ]


class NamespaceManagementDataStructure(Structure):
    _pack_ = 1
    _fields_ = [
        ('NSZE', c_uint64),
        ('NCAP', c_uint64),
        ('RSV0', c_uint8 * 10),
        ('FLBAS', c_uint8),
        ('RSV1', c_uint16),
        ('DPS', c_uint8),
        ('NMIC', c_uint8),
        ('RSV2', c_uint8 * 4065)
    ]


class DatasetManagementContextAttributes(Structure):
    _pack_ = 1
    _fields_ = [
        ('AF', c_uint32, 4),
        ('AL', c_uint32, 2),
        ('RSV0', c_uint32, 2),
        ('SR', c_uint32, 1),
        ('SW', c_uint32, 1),
        ('WP', c_uint32, 1),
        ('RSV1', c_uint32, 13),
        ('CAS', c_uint32, 8),
    ]

class DebugInfo(Structure):
    _pack_ = 1
    _fields_ = [
        ('PHASE', c_uint8),
        ('MODE', c_uint8),
        ('RSV', c_uint16),
        ('LBA', c_uint32),
        ('KEY', c_uint32),
    ]


class AsyncTrace(Structure):
    _pack_ = 1
    _fields_ = [
        ('entry', DebugInfo*5000),
    ]


class DatasetManagementEntry(Structure):
    _pack_ = 1
    _fields_ = [
        ('DMCA', DatasetManagementContextAttributes),
        ('LENGTH', c_uint32),
        ('SLBA', c_uint64),
    ]


class SglDescriptorFormat(Structure):
    """SGL Descriptor Format"""
    _pack_ = 1
    _fields_ = [
        ('address', c_uint64),
        ('length', c_uint64, 32),
        ('rsv1', c_uint64, 24),
        ('dts', c_uint64, 4),       # Descriptor Type Specific
        ('sdt', c_uint64, 4)        # SGL Descriptor Type
    ]


class LiveMigrationSQE(Structure):
    _pack_ = 1
    _fields_ = [
        ('OPCODE', c_uint32, 8),
        ('FUSE', c_uint32, 2),
        ('RSV0', c_uint32, 4),
        ('PSDT', c_uint32, 2),
        ('CID', c_uint32, 16),
        ('CNTID', c_uint16, 16),
        ('RSV0', c_uint16, 16),
        ('BK_HEAD', c_uint64, 64),
        ('HEAD', c_uint64, 64),
        ('TAIL', c_uint64, 64),
        ('PRP', c_uint64, 64),
        ('CDW10_OPCODE', c_uint32, 8),
        ('CDW10_RSV', c_uint32, 24),
        ('NPAGES', c_uint32, 32),
        ('CDW12', c_uint32, 32),
        ('CDW13', c_uint32, 32),
        ('CDW14', c_uint32, 32),
        ('SUB_OPCODE', c_uint32, 32),
        # ('TIMEOUT', c_uint32),
        # ('RESULT', c_uint32),
    ]


class LVMTrackIOEntry(Structure):
    _pack_ = 1
    _fields_ = [
        ('TYPE', c_uint8, 1),
        ('NSID', c_uint8, 7),
        ('NLB', c_uint16, 16),
        ('SLBA_L', c_uint32),
        ('SLBA_H', c_uint8)

    ]

STATUS_FIELD = {
    0x0: 'Successful Completion',
    0x1: 'Invalid Command Opcode',
    0x2: 'Invalid Field in Command',
    0x3: 'Command ID Conflict',
    0x4: 'Data Transfer Error',
    0x5: 'Commands Aborted due to Power Loss Notification',
    0x6: 'Internal Error',
    0x7: 'Command Abort Requested',
    0x8: 'Command Aborted due to SQ Deletion',
    0x9: 'Command Aborted due to Failed Fused Command',
    0xa: 'Command Aborted due to Missing Fused Command',
    0xb: 'Invalid Namespace or Format',
    0xc: 'Command Sequence Error',
    0xd: 'Invalid SGL Segment Descriptor',
    0xe: 'Invalid Number of SGL Descriptors',
    0xf: 'Data SGL Length Invalid',
    0x10: 'Metadata SGL Length Invalid',
    0x11: 'SGL Descriptor Type Invalid',
    0x12: 'Invalid Use of Controller Memory Buffer',
    0x13: 'PRP Offset Invalid',
    0x14: 'Atomic Write Unit Exceeded',
    0x15: 'Operation Denied',
    0x16: 'SGL Offset Invalid',
    0x18: 'Host Identifier Inconsistent Format',
    0x19: 'Keep Alive Timeout Expired',
    0x1a: 'Keep Alive Timeout Invalid',
    0x1b: 'Command Aborted due to Preempt and Abort',
    0x1c: 'Sanitize Failed',
    0x1d: 'Sanitize In Progress',
    0x1e: 'SGL Data Block Granularity Invalid',
    0x1f: 'Command Not Supported for Queue in CMB',
    0x80: 'LBA Out of Range',
    0x81: 'Capacity Exceeded',
    0x82: 'Namespace Not Ready',
    0x83: 'Reservation Conflict',
    0x84: 'Format In Progress',

    0x100: 'Completion Queue Invalid',
    0x101: 'Invalid Queue Identifier',
    0x102: 'Invalid Queue Size',
    0x103: 'Abort Command Limit Exceeded',
    0x105: 'Asynchronous Event Request Limit Exceeded',
    0x106: 'Invalid Firmware Slot',
    0x107: 'Invalid Firmware Image',
    0x108: 'Invalid Interrupt Vector',
    0x109: 'Invalid Log Page',
    0x10a: 'Invalid Format',
    0x10b: 'Firmware Activation Requires Conventional Reset',
    0x10c: 'Invalid Queue Deletion',
    0x10d: 'Feature Identifier Not Saveable',
    0x10e: 'Feature Not Changeable',
    0x10f: 'Feature Not Namespace Specific',
    0x110: 'Firmware Activation Requires NVM Subsystem Reset',
    0x111: 'Firmware Activation Requires Controller Level Reset',
    0x112: 'Firmware Activation Requires Maximum Time Violation',
    0x113: 'Firmware Activation Prohibited',
    0x114: 'Overlapping Range',
    0x115: 'Namespace Insufficient Capacity',
    0x116: 'Namespace Identifier Unavailable',
    0x118: 'Namespace Already Attached',
    0x119: 'Namespace Is Private',
    0x11a: 'Namespace Not Attached',
    0x11b: 'Thin Provisioning Not Supported',
    0x11c: 'Controller List Invalid',
    0x11d: 'Device Self-test In Progress',
    0x11e: 'Boot Partition Write Prohibited',
    0x11f: 'Invalid Controller Identifier',
    0x120: 'Invalid Secondary Controller State',
    0x121: 'Invalid Number of Controller Resources',
    0x122: 'Invalid Resource Identifier',
    0x180: 'Conflicting Attributes',
    0x181: 'Invalid Protection Information',
    0x182: 'Attempted Write to Read Only Range',

    0x280: 'Write Fault',
    0x281: 'Unrecovered Read Error',
    0x282: 'End-to-end Guard Check Error',
    0x283: 'End-to-end Application Tag Check Error',
    0x284: 'End-to-end Reference Tag Check Error',
    0x285: 'Compare Failure',
    0x286: 'Access Denied',
    0x287: 'Deallocated or Unwritten Logical Block',
}



from utils import log

class Message(object):

    def __init__(self, msg=""):
        self.msg = str()
        self.add_message(msg)

    def add_message(self, message):
        log.INFO(message)
        self.msg = "{}\n{}".format(self.msg, message)

    @property
    def message(self):
        return self.msg

    def clear_message(self):
        self.msg = str()


import os
from utils import log
import time
import subprocess
from test_framework.state import State


class NoseEngine(object):

    def __init__(self):
        self.working_path = os.environ["working_path"]
        self.log_path = self.get_log_path()

    def get_log_path(self):
        log_path = os.path.join(self.working_path, "log")
        if os.path.exists(log_path) is False:
            os.mkdir(log_path)
        return log_path

    @staticmethod
    def get_msg(out, err):
        msg = str()
        if err is not None:
            msg += err.decode('utf-8')
        if out is not None:
            msg += out.decode('utf-8')
        return msg

    def save_msg(self, msg, test_case):
        test_function = test_case.split(".")
        log_file = "{}_{}.log".format(test_function[-1],time.time())
        log_path = os.path.join(self.log_path, log_file)
        with open(log_path, "w") as file_:
            file_.write(msg)
        return log_path

    def command_run_test(self, test_case, test_path):
        test_function = test_case.split(".")
        xml_name = "nosetests_%s_%s.xml" % (test_function[-1], time.time())
        xml_path = os.path.join(self.log_path, xml_name)
        command_line = "nosetests --exe --nocapture --with-printlog --with-xunit --xunit-file={} -x {}"\
            .format(xml_path, test_path)
        child1 = subprocess.Popen(command_line, shell=True)
        return_code = child1.wait()
        ret = True if return_code == 0 else False
        return ret, xml_path

    def run(self, test_case, test_path, parameters, queue):
        ret,  xml_path = self.command_run_test(test_case, test_path)
        if ret is True:
            log.INFO("TestCase run succeed.%s", test_case)
            result = State.PASS
        else:
            log.ERR("TestCase run failed. %s", test_case)
            result = State.FAIL
        result = {"name":test_case, "result": result, "log":xml_path, "xml_path": xml_path}
        queue.put(result)
        return ret



from test_framework.state import State, DownloadType
from test_framework.upgrade_engine.serial_download import SerialDownloader
from utils.process import MyProcess
from multiprocessing import Queue


class FirmwareDownloader(object):

    def __init__(self, execute_name):
        self.process_run_ = None
        self.results = list()
        self.execute_name = execute_name
        self.runner = None

    def get_download_engine(self, test_parameters):
        if test_parameters["download_type"] == DownloadType.SerialTestPlatform:
            self.runner = SerialDownloader()
        else:
            self.runner = None

    def _run(self, test_parameters, queue):
        try:
            print("FIRMDOWNLOAD", test_parameters)
            self.get_download_engine(test_parameters)
            status, log = self.runner.run(test_parameters)
            ret = State.PASS if status == 0 else State.FAIL
            result = {"name": self.execute_name, "result": ret, "log":log}
            queue.put(result)
        except Exception as e:
            print(e)
            result = None
        return result

    def run(self, test_parameters):
        print(test_parameters)
        queue = Queue()
        self.process_run_ = MyProcess(target=self._run, args=(test_parameters, queue, ))
        self.process_run_.start()
        self.process_run_.join()
        value = queue.get(True)
        self.results.append(value)
        return self.results

    def stop(self):
        print("FirmwareDownloader runner . stop")
        if self.process_run_ is not None:
            ret = self.process_run_.stop()
        else:
            ret = -1
        return ret


# pylint: disable=broad-except, invalid-name
import os
from utils import log
from test_framework.database.database import SqlConnection
from test_framework.state import SlotState
from utils.system import get_host_name, get_multi_ip_address


class DutDatabase(SqlConnection):

    def __init__(self):
        super(DutDatabase, self).__init__(db_name="dutdb")
        self.ip_address = get_multi_ip_address()
        self.port = os.environ.get('agent_port', '5000')
        self.agent_table = "agent"
        self.slot_table = "slot"

    def update_agent(self, agent):
        agent_id = self.get_agent_id(self.ip_address, self.port)
        if agent_id is not None:
            self.update_exist_agent(agent, agent_id)
        else:
            self.create_new_agent(agent)

    def update_slot(self, slot_id, slot):
        if slot_id != -1:
            self.update_exist_slot(slot_id, slot)
        else:
            self.create_new_slot(slot)

    def update_exist_slot(self, slot_id, slot):
        update_str = self._covert_dict_2_update_string(**slot)
        update_command = "UPDATE {} SET {} where `id`='{}'".format(self.slot_table, update_str, slot_id)
        self.cursor.execute(update_command)
        self.conn.commit()

    def create_new_slot(self, slot):
        col_str, value_str = self._covert_dict_2_insert_string(**slot)
        insert_command = "INSERT INTO `{}` ({}) VALUES({})".format(self.slot_table, col_str, value_str)
        self.cursor.execute(insert_command)
        self.conn.commit()

    def is_exist_agent(self, ip_addresses, port):
        result = False
        for ip in ip_addresses:
            sql_command = "SELECT * from {} WHERE ip like'%{}%' AND port='{}'".format(self.agent_table, ip, port)
            gets = self.execute_sql_command(sql_command)
            if gets:
                result = True
                break
        return result

    def get_agent_id(self, ip_addresses, port):
        agent_id = None
        for ip in ip_addresses:
            sql_command = "SELECT * from {} WHERE ip like'%{}%' AND port='{}'".format(self.agent_table, ip, port)
            gets = self.execute_sql_command(sql_command)
            if gets:
                agent_id = gets[0][0]
                break
        return agent_id

    def is_exist_slot(self, ogt_config, agent_id):
        sql_command = "SELECT * from {} WHERE config_name='{}' AND agent='{}'".format(self.slot_table, ogt_config, agent_id)
        gets = self.execute_sql_command(sql_command)
        result = True if gets else False
        return result

    def update_exist_agent(self, agent, agent_id):
        str_date = "`os`='{}',`platform`='{}',`ip`='{}',`name`='{}'".format(agent["os"], agent["platform"], agent["ip"],
                                                                            agent["name"])
        cmd = "UPDATE {} SET {} WHERE `id`='{}'".format(self.agent_table, str_date, agent_id)
        self.cursor.execute(cmd)
        self.conn.commit()

    def create_new_agent(self, agent):
        self.insert_to_table(self.agent_table,
                             name=agent["name"],
                             ip=agent["ip"],
                             port=agent["port"],
                             os=agent["os"],
                             platform=agent["platform"])

    def get_agent_related_slots(self, agent):
        cmd = "SELECT * FROM {} where agent='{}'".format(self.slot_table, agent)
        result = self.execute_sql_command(cmd)
        return result

    def get_agent_related_slots_by_id(self, agent_id):
        cmd = "SELECT * FROM {} where `agent`='{}'".format(self.slot_table, agent_id)
        result = self.execute_sql_command(cmd)
        return result

    def agent_heart_beat(self, state):
        update_time = self.get_datetime()
        agent_id = self.get_agent_id(self.ip_address, self.port)
        status = "`status`='{}', update_time='{}'".format(state, update_time)
        update_command = "UPDATE {} SET {} WHERE `id`='{}'".format(self.agent_table, status, agent_id)
        self.cursor.execute(update_command)
        self.conn.commit()

    def delete_removed_slot(self, slot_id):
        if slot_id is not None:
            update_command = "DELETE FROM {} WHERE `id`='{}'".format(self.slot_table, slot_id)
            self.cursor.execute(update_command)
            self.conn.commit()

    def set_slot_to_lost(self, slot_id):
        if slot_id is not None:
            log.INFO("Find slot is lost: {}".format(slot_id))
            cmd = "UPDATE  {} SET `status`='{}' WHERE `id`='{}'".format(self.slot_table, SlotState.Lost, slot_id)
            self.cursor.execute(cmd)
            self.conn.commit()

    def get_all_agents(self):
        sql_cmd = "SELECT * from {} ".format(self.agent_table)
        agents = self.execute_sql_command(sql_cmd)
        return agents

    def get_all_slots(self):
        sql_cmd = "SELECT * from {} ".format(self.slot_table)
        slots = self.execute_sql_command(sql_cmd)
        return slots

    @staticmethod
    def map_agent_name_2_id(agent_name, all_agents):
        agent_id = None
        for agent in all_agents:
            if agent[1] == agent_name:
                agent_id = agent[0]
                break
        return agent_id

    def update_slot_agent_2_id(self):
        agents = self.get_all_agents()
        slots = self.get_all_slots()
        for slot in slots:
            agent_name = slot[15]
            agent_id = self.map_agent_name_2_id(agent_name, agents)
            if agent_id is not None:
                cmd = "UPDATE  {} SET `agent`='{}' WHERE `id`='{}'".format(self.slot_table, agent_id, slot[0])
                self.cursor.execute(cmd)
                self.conn.commit()

    def set_ogt_config_2_slot(self, configs):
        for config in configs:
            ip = config["ip"]
            ogt_config = config["ogt_config"]
            ip_list = list()
            ip_list.append(ip)
            agent_id = self.get_agent_id(ip_list, "5000")
            if agent_id is None:
                agent = {"name": "", "ip": ip, "port": "5000", "os": "windows", "platform": "oakgate"}
                self.create_new_agent(agent)
                agent_id = self.get_agent_id(ip_list, "5000")
            if self.is_exist_slot(ogt_config, agent_id) is False:
                slot = {"agent": agent_id, "config_name": ogt_config}
                self.create_new_slot(slot)


def update_agent(func):
    def func_wrapper(*args, **kwargs):
        agent = func(*args, **kwargs)
        try:
            sql_connection = DutDatabase()
            sql_connection.update_agent(agent)
        except Exception as all_exception:
            log.ERR(all_exception)
        return agent
    return func_wrapper


def update_slot(func):
    def func_wrapper(*args, **kwargs):
        slot_id, slot = func(*args, **kwargs)
        if slot is not None:
            try:
                sql_connection = DutDatabase()
                sql_connection.update_slot(slot_id, slot)
            except Exception as all_exception:
                log.ERR(all_exception)
        return slot_id, slot
    return func_wrapper


def agent_heart_beat(func):
    def func_wrapper(*args, **kwargs):
        status = func(*args, **kwargs)
        try:
            sql_connection = DutDatabase()
            sql_connection.agent_heart_beat(status)
        except Exception as all_exception:
            log.ERR(all_exception)
        return status
    return func_wrapper


def delete_removed_slot(func):
    def func_wrapper(*args, **kwargs):
        slot_id = func(*args, **kwargs)
        try:
            sql_connection = DutDatabase()
            sql_connection.delete_removed_slot(slot_id)
        except Exception as all_exception:
            log.ERR(all_exception)
        return slot_id
    return func_wrapper


def check_lost_slot(func):
    def func_wrapper(*args, **kwargs):
        slot_id = func(*args, **kwargs)
        try:
            sql_connection = DutDatabase()
            sql_connection.set_slot_to_lost(slot_id)
        except Exception as all_exception:
            log.ERR(all_exception)
        return slot_id
    return func_wrapper



#!/usr/bin/env python
# pylint: disable=too-many-locals,too-many-arguments,superfluous-parens,import-error,invalid-name,arguments-differ
"""
Created on 2019/4/26

@author: yyang
"""
from ctypes import *
from tool.device.nvme.struct import *
from utils import log
from tool.device.nvme.buf import *


class VUProjectType:

    def __init__(self):
        pass

    @classmethod
    def from_name(cls, name):
        """
            Return the class which matches ``name`` if it exists inside this
            module or raise an exception.
        """
        if name not in globals():
            raise NotImplementedError('Project of type %s is not supported' % name)
        return globals()[name]


class TahoeVU(VUProjectType):
    def __init__(self, dev):
        super().__init__()
        self.dev = dev
        log.INFO("Tahoe VU class init!")

    def malloc(self, size):
        ptr = self.dev.dll.nvme_alloc(size)
        assert ptr != 0, 'Malloc {:#x} Failed!'.format(size)
        return ptr

    def show_test_case(self, show_string):
        """Show test case in UART log
        Domain      DW10       DW11       DW12   DW13    DW14  DW15   DW0(CQE)
        | Media  |  0x10001007  |  1024 |  None |  None  | None | None| None

        This is VU command for showing test case in UART log
        Args:
           None
        Returns:
           None
        Raises:
           send VU cmd fail
        """
        str1 = show_string.encode()
        size = 512  # size must be 512, otherwise standard driver can not allocate memory
        dptr = create_string_buffer(str1, size)
        ptr = self.malloc(size)
        try:
            self.dev.dll.nvme_copy(c_void_p(ptr), dptr, size)

            cmd = NVMePassthruCmd()
            cmd.OPCODE = 0xC0
            cmd.CDW[10] = 0x10001007
            cmd.CDW[11] = 1024
            cmd.DATA = ptr
            cmd.DATA_LEN = size
            #log.INFO("%x", ptr)
            #log.INFO(list(dptr))
            return self.dev.admin_passthru(cmd)
        finally:
            self.dev.dll.nvme_free(c_void_p(ptr))

    def get_power_consumption(self, length=64):
        """get_async_event_info

            Domain         DW10       DW11         DW12               DW13    DW14   DW15  DW0(CQE)
        |Host/Platform| 0x10002028  | 4    |        None            | None  | None  | None| None

        Args:
            None
        Returns:
            power consumption data: Word0: Voltage, Word1: Current
        """
        dptr = TahoeVUPower()
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x10002028
        cmd.CDW[11] = length
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        vol = dptr.vol
        cur = dptr.cur
        ret = float(str(vol)+'.'+str(cur))
        #log.INFO("Get Power Consumption syccessfully[int: %x, float: %x].", vol, cur)
        log.INFO("Get Power Consumption syccessfully[Power: %f].", ret)
        log.INFO("Get Power Consumption syccessfully[Credit: max:%d, min:%d, avg:%d].",\
                 dptr.max_credit, dptr.min_credit, dptr.avg_credit)
        return ret, dptr.max_credit, dptr.min_credit, dptr.avg_credit

    def get_memory(self, addr, length=1, cpu=0):
        dptr = (c_uint32 * length)()
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x10002005
        cmd.CDW[11] = length
        cmd.CDW[12] = cpu
        cmd.CDW[13] = addr
        cmd.CDW[14] = length * 4
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        return dptr

    def set_memory(self, addr, dptr, length=0, cpu=0):
        ptr = self.malloc(sizeof(dptr))

        try:
            length = length if length else sizeof(dptr) // 4
            self.dev.dll.nvme_copy(c_void_p(ptr), dptr, length * 4)

            cmd = NVMePassthruCmd()
            cmd.OPCODE = 0xC0
            cmd.CDW[10] = 0x10001006
            cmd.CDW[11] = length
            cmd.CDW[12] = cpu
            cmd.CDW[13] = addr
            cmd.CDW[14] = length * 4
            cmd.DATA = ptr
            cmd.DATA_LEN = sizeof(dptr)
            return self.dev.admin_passthru(cmd)
        finally:
            self.dev.dll.nvme_free(c_void_p(ptr))

    def get_idle_state(self):

        #dptr = (c_uint16 * (length//2))()
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x10002042
        cmd.CDW[11] = 1
        #cmd.DATA = addressof(dptr)
        #cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        return cmd.RESULT


    def force_gc(self, blk):

        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x2000000C
        cmd.CDW[12] = blk

        return self.dev.admin_passthru(cmd)

    def get_bmi_entry(self, blk):
        dptr = TahoeBMIEntry()
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x20002001
        cmd.CDW[11] = 32
        cmd.CDW[12] = 1
        cmd.CDW[13] = blk

        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        return dptr

    def force_flush(self, task="host", ftype="block"):
        cmd = NVMePassthruCmd()
        assert task in ["host", "gc", "ftl"], "task type not support!"
        assert ftype in ["raidline", "block", "pu", "pad", "wordline"], \
                "task type not support!"
        if task == "host":
            cmd.CDW[12] = 0
        elif task == "gc":
            cmd.CDW[12] = 1
        elif task == "ftl":
            cmd.CDW[12] = 2

        if ftype == "raidline":
            cmd.CDW[13] = 0
        elif ftype == "block":
            cmd.CDW[13] = 1
        elif ftype == "pu":
            cmd.CDW[13] = 2
        elif ftype == "pad":
            cmd.CDW[13] = 3
        elif ftype == "wordline":
            cmd.CDW[13] = 4
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x20000011

        return self.dev.admin_passthru(cmd)

    def get_int_type(self):
        """Get INT type
        Domain      DW10          DW11    DW12    DW13    DW14  DW15   DW0(CQE)
        | Media  |  0x10002034   |  4   |  None |  None  | None | None |  None
        Args:
            None
        Returns:
            INT type:INT0/INT1 or INT2
        Raises:
            send VU cmd fail
        """
        dptr = (c_uint8 * (32))()
        cmd = NVMePassthruCmd()
        dw10 = 0x10002034
        cmd.CDW[10] = dw10
        cmd.CDW[11] = 4
        cmd.OPCODE = 0xC0
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        tmp = map(chr, dptr[0:17])
        int_type = "".join(tmp)
        int_type = int_type.strip('\x00')
        log.INFO("INT type is %s", int_type)
        return int_type

    def get_cmd_cnt_per_queue(self, qid=1):
        """get_cmd_cnt_per_queue

            Domain         DW10       DW11    DW12   DW13    DW14   DW15  DW0(CQE)
        |Host/Platform| 0x10000043    | 0x0 |  qid |     None | None | None| None

        This is VU command for get command counter per queu
        Args:
            None
        Returns:
            cqe dw0: command counter per Queue
        """
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x10000043
        cmd.CDW[12] = qid
        #cmd.DATA = addressof(dptr)
        #cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        log.INFO("get_cmd_cnt_per_queue[q:%d, number:%d] successfully.", qid, cmd.RESULT)
        return cmd.RESULT

    def set_crecit(self, val):
        cmd = NVMePassthruCmd()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = 0x10000046
        cmd.CDW[11] = 0
        cmd.CDW[12] = int(val)

        return self.dev.admin_passthru(cmd)

    def get_drive_security_state(self):
        """this function is getting drive security state  (ISE/SED/TCG not enable)

                Domain         DW10        DW11       DW12    DW13    DW14  DW15   DW0(CQE)
               | Generic   |  0x1000204f  |  0x04  |  0    |   0    |   0| None

               This is VU command for getting drive security state
               Args:
                   None
               Returns:
                   Returns TCG State:
                    byte 0 - 1: return data length (big-endian)
                    byte 2: 0 = SED, 1 = ISE
                    byte 3: 1 = TCG is finalized
               Raises:
                   send VU cmd fail
               """
        cmd = NVMePassthruCmd()
        dw10 = 0x1000204f
        dptr = (c_uint8 * 4096)()
        cmd.OPCODE = 0xC0
        cmd.CDW[10] = dw10
        cmd.CDW[11] = 0x04
        cmd.DATA = addressof(dptr)
        cmd.DATA_LEN = sizeof(dptr)
        self.dev.admin_passthru(cmd)
        return dptr



# coding=utf-8
import os
from test_framework.test_runner import Runner
from utils import log
from test_framework.test_base import TestBase


class TestCase(TestBase):

    def __init__(self, test_case_name=None):
        super(TestCase, self).__init__()
        self.test_name = test_case_name
        self.runner = Runner()
        self.get_all_modules(self.test_case_path)
        self.list_tests()

    def stop(self):
        return self.runner.stop()

    def get_test_cases(self):
        return self.tests

    def list_and_filter_tests(self, filters):
        filter_result = []
        if filters.lower() == "all":
            filter_result = self.tests
        else:
            for item in self.tests:
                if filters in item:
                    filter_result.append(item)
        return filter_result

    def run(self, test_name=None, parameters=None):
        if test_name is not None:
            self.test_name = test_name
        test_path = self.get_all_script_path(self.test_name)
        if not test_path:
            log.ERR("TestCase not find: %s", self.test_name)
        else:
            self.runner.process_run(self.test_name, test_path[0], parameters, 1, 0)
        return self.runner.get_results()



# coding=utf-8
# pylint: disable=redefined-builtin
import os
import threading
from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer
from pyftpdlib.authorizers import DummyAuthorizer
from utils.system import get_root_path


class MyHandler(FTPHandler):

    def on_connect(self):
        print("%s:%s connected" % (self.remote_ip, self.remote_port))

    def on_disconnect(self):
        # do something when client disconnects
        pass

    def on_login(self, username):
        # do something when user login
        pass

    def on_logout(self, username):
        # do something when user logs out
        pass

    def on_file_sent(self, file):
        # do something when a file has been sent
        print(self.username, file)

    def on_file_received(self, file):
        # do something when a file has been received
        print(self.username, file)

    def on_incomplete_file_sent(self, file):
        # do something when a file is partially sent
        print(self.username, file)

    def on_incomplete_file_received(self, file):
        # remove partially uploaded files
        os.remove(file)


def start_ftp_server():
    working_path = os.environ["working_path"]
    home_dir = os.path.join(working_path, "fw_bin")
    if os.path.exists(home_dir) is False:
        os.mkdir(home_dir)
    authorizer = DummyAuthorizer()
    authorizer.add_user('ftp_user', 'Cnex!321', homedir=home_dir, perm='elw')
    authorizer.add_anonymous(homedir=home_dir)
    handler = MyHandler
    handler.authorizer = authorizer
    try:
        server = FTPServer(('0.0.0.0', 10021), handler)
        server.serve_forever()
        print("FTP server start succeed.")
    except Exception as all_exception:
        print("FTP server already start.")

def thread_start_ftp_server():
    thread_p = threading.Thread(target=start_ftp_server)
    thread_p.setDaemon(True)
    thread_p.start()


if __name__ == "__main__":
    thread_start_ftp_server()



#!/usr/bin/env python
# pylint: disable=superfluous-parens,missing-docstring,too-few-public-methods
"""
Created on 2020/06/10

@author: yyang
"""
from ctypes import Structure, c_uint8, c_uint16, c_uint32


class ID(Structure):
    _pack_ = 1
    _fields_ = [
        ('VID', c_uint32, 16),  # Vendor ID (VID)
        ('DID', c_uint32, 16),  # Device ID (DID)
    ]
    desc = {'VID': 'Vendor ID', 'DID': 'Device ID'}


class CMD(Structure):
    _pack_ = 1
    _fields_ = [
        ('IOSE', c_uint16, 1),  # I/O Space Enable (IOSE)
        ('MSE', c_uint16, 1),  # Memory Space Enable (MSE)
        ('BME', c_uint16, 1),  # Bus Master Enable (BME)
        ('SCE', c_uint16, 1),  # Special Cycle Enable (SCE)
        ('MWIE', c_uint16, 1),  # Memory Write and Invalidate Enable (MWIE)
        ('VGA', c_uint16, 1),  # VGA Palette Snooping Enable (VGA)
        ('PEE', c_uint16, 1),  # Parity Error Response Enable (PEE)
        ('HW', c_uint16, 1),  # Hardwired
        ('SEE', c_uint16, 1),  # SERR# Enable (SEE)
        ('FBE', c_uint16, 1),  # Fast Back-to-Back Enable (FBE)
        ('ID', c_uint16, 1),  # Interrupt Disable (ID)
        ('RSV', c_uint16, 5),
    ]
    desc = {
        'IOSE': 'I/O Space Enable', 'MSE': 'Memory Space Enable', 'BME': 'Bus Master Enable',
        'SCE': 'Special Cycle Enable', 'MWIE': 'Memory Write and Invalidate Enable',
        'VGA': 'VGA Palette Snooping Enable', 'PEE': 'Parity Error Response Enable', 'HW': 'Hardwired',
        'SEE': 'SERR# Enable', 'FBE': 'Fast Back-to-Back Enable', 'ID': 'Interrupt Disable'
    }


class STS(Structure):
    _pack_ = 1
    _fields_ = [
        ('RSV0', c_uint16, 3),
        ('IS', c_uint16, 1),  # Interrupt Status (IS)
        ('CL', c_uint16, 1),  # Capabilities List (CL)
        ('C66', c_uint16, 1),  # 66 MHz Capable (C66)
        ('RSV1', c_uint16, 1),
        ('FBC', c_uint16, 1),  # Fast Back-to-Back Capable (FBC)
        ('DPD', c_uint16, 1),  # Master Data Pariy Error Detected (DPD)
        ('DEVT', c_uint16, 2),  # DEVSEL# Timing (DEVT)
        ('STA', c_uint16, 1),  # Signaled Target-Abort (STA)
        ('RTA', c_uint16, 1),  # Received Target Abort (RTA)
        ('RMA', c_uint16, 1),  # Received Master-Abort (RMA)
        ('SSE', c_uint16, 1),  # Signaled System Error (SSE)
        ('DPE', c_uint16, 1),  # Detected Parity Error (DPE)
    ]
    desc = {
        'IS': 'Interrupt Status', 'CL': 'Capabilities List', 'C66': '66 MHz Capable',
        'FBC': 'Fast Back-to-Back Capable', 'DPD': 'Master Data Pariy Error Detected', 'DEVT': 'DEVSEL# Timing',
        'STA': 'Signaled Target-Abort', 'RTA': 'Received Target Abort', 'RMA': 'Received Master-Abort',
        'SSE': 'Signaled System Error', 'DPE': 'Detected Parity Error',
    }


class CC(Structure):
    _pack_ = 1
    _fields_ = [
        ('PI', c_uint8),  # Programming Interface (PI)
        ('SCC', c_uint8),  # Sub Class Code (SCC)
        ('BCC', c_uint8),  # Base Class Code (BCC)
    ]
    desc = {'PI': 'Programming Interface', 'SCC': 'Sub Class Code', 'BCC': 'Base Class Code'}


class HTYPE(Structure):
    _pack_ = 1
    _fields_ = [
        ('HL', c_uint8, 7),  # Header Layout (HL)
        ('MFD', c_uint8, 1),  # Multi-Function Device (MFD)
    ]
    desc = {'HL': 'Header Layout', 'MFD': 'Multi-Function Device'}


class BIST(Structure):
    _pack_ = 1
    _fields_ = [
        ('CC', c_uint8, 4),  # Completion Code (CC)
        ('RSV', c_uint8, 2),
        ('SB', c_uint8, 1),  # Start BIST (SB)
        ('BC', c_uint8, 1),  # BIST Capable (BC)
    ]
    desc = {'CC': 'Completion Code', 'SB': 'Start BIST', 'BC': 'BIST Capable'}


class MLBAR(Structure):
    _pack_ = 1
    _fields_ = [
        ('RTE', c_uint32, 1),  # Resource Type Indicator (RTE)
        ('TP', c_uint32, 2),  # Type (TP)
        ('PF', c_uint32, 1),  # Prefetchable (PF)
        ('RSV', c_uint32, 10),
        ('BA', c_uint32, 18),  # Base Address (BA)
    ]
    desc = {'RTE': 'Resource Type Indicator', 'TP': 'Type', 'PF': 'Prefetchable', 'BA': 'Base Address'}


class BAR(Structure):
    _pack_ = 1
    _fields_ = [
        ('RTE', c_uint32, 1),  # Resource Type Indicator (RTE)
        ('RSV', c_uint32, 2),
        ('BA', c_uint32, 29),  # Base Address (BA)
    ]
    desc = {'RTE': 'Resource Type Indicator', 'BA': 'Base Address'}


class SS(Structure):
    _pack_ = 1
    _fields_ = [
        ('SSVID', c_uint32, 16),  # Subsystem Vendor ID (SSVID)
        ('SSID', c_uint32, 16),  # Subsystem ID (SSID)
    ]
    desc = {'SSVID': 'Subsystem Vendor ID', 'SSID': 'Subsystem ID'}


class INTR(Structure):
    _pack_ = 1
    _fields_ = [
        ('ILINE', c_uint16, 8),  # Interrupt Line (ILINE)
        ('IPIN', c_uint16, 8),  # Interrupt Pin (IPIN)
    ]
    desc = {'ILINE': 'Interrupt Line', 'IPIN': 'Interrupt Pin'}


class PCIHeader(Structure):
    """System Bus (PCI Express) Registers - PCI Header"""
    _pack_ = 1
    _fields_ = [
        ('ID', ID),  # Identifiers
        ('CMD', CMD),  # Command Register
        ('STS', STS),  # Device Status
        ('RID', c_uint8),  # Revision ID
        ('CC', CC),  # Class Codes
        ('CLS', c_uint8),  # Cache Line Size
        ('MLT', c_uint8),  # Master Latency Timer
        ('HTYPE', HTYPE),  # Header Type
        ('BIST', BIST),  # Built In Self Test (Optional)
        ('MLBAR', MLBAR),  # Memory Register Base Address, lower 32-bits <BAR0>
        ('MUBAR', c_uint32),  # Memory Register Base Address, upper 32-bits <BAR1>
        ('BAR2', BAR),  # Refer to section 2.1.12
        ('BAR3', c_uint32),  # Vendor Specific
        ('BAR4', c_uint32),  # Vendor Specific
        ('BAR5', c_uint32),  # Vendor Specific
        ('CCPTR', c_uint32),  # CardBus CIS Pointer
        ('SS', SS),  # Subsystem Identifiers
        ('EROM', c_uint32),  # Expansion ROM Base Address (Optional)
        ('CAP', c_uint8),  # Capabilities Pointer
        ('RSV', c_uint8 * 7),  #
        ('INTR', INTR),  # Interrupt Information
        ('MGNT', c_uint8),  # Minimum Grant (Optional)
        ('MLAT', c_uint8),  # Maximum Latency (Optional)
    ]
    desc = {
        'RID': 'Revision ID', 'CLS': 'Cache Line Size', 'MLT': 'Master Latency Timer', 'MUBAR': 'BAR1 Address',
        'BAR3': 'BAR3 Address', 'BAR4': 'BAR4 Address', 'BAR5': 'BAR5 Address', 'CCPTR': 'CardBus CIS Pointer',
        'EROM': 'Expansion ROM Base Address', 'CAP': 'Capabilities Pointer', 'MGNT': 'Minimum Grant',
        'MLAT': 'Maximum Latency',
    }


class PID(Structure):
    _pack_ = 1
    _fields_ = [
        ('CID', c_uint16, 8),  # Cap ID (CID)
        ('NEXT', c_uint16, 8),  # Next Capability (NEXT)
    ]
    desc = {'CID': 'Capability ID', 'NEXT': 'Next Capability'}


class PC(Structure):
    _pack_ = 1
    _fields_ = [
        ('VS', c_uint16, 3),  # Version (VS)
        ('PMEC', c_uint16, 1),  # PME Clock (PMEC)
        ('RSV', c_uint16, 1),
        ('DSI', c_uint16, 1),  # Device Specific Initialization (DSI)
        ('AUXC', c_uint16, 3),  # Aux_Current (AUXC)
        ('D1S', c_uint16, 1),  # D1_Support (D1S)
        ('D2S', c_uint16, 1),  # D2_Support (D2S)
        ('PSUP', c_uint16, 5),  # PME_Support (PSUP)
    ]
    desc = {
        'VS': 'Version', 'PMEC': 'PME Clock', 'DSI': 'Device Specific Initialization',
        'AUXC': 'Aux_Current', 'D1S': 'D1_Support', 'D2S': 'D2_Support', 'PSUP': 'PME_Support'
    }


class PMCS(Structure):
    _pack_ = 1
    _fields_ = [
        ('PS', c_uint16, 2),  # Power State (PS)
        ('RSV1', c_uint16, 1),
        ('NSFRST', c_uint16, 1),  # No Soft Reset (NSFRST)
        ('RSV2', c_uint16, 4),
        ('PMEE', c_uint16, 1),  # PME Enable (PMEE)
        ('DSE', c_uint16, 4),  # Data Select (DSE)
        ('DSC', c_uint16, 2),  # Data Scale (DSC)
        ('PMES', c_uint16, 1),  # PME Status (PMES)
    ]
    desc = {
        'PS': 'Power State', 'NSFRST': 'No Soft Reset', 'PMEE': 'PME Enable',
        'DSE': 'Data Select', 'DSC': 'Data Scale', 'PMES': 'PME Status'
    }


class PMCAP(Structure):
    """System Bus (PCI Express) Registers - PCI Power Management Capabilities"""
    _pack_ = 1
    _fields_ = [
        ('PID', PID),  # PCI Power Management Capability ID
        ('PC', PC),  # PCI Power Management Capabilities
        ('PMCS', PMCS),  # PCI Power Management Control and Status
    ]
    desc = {
        'PID': 'PCI Power Management Capability ID', 'PC': 'PCI Power Management Capabilities',
        'PMCS': 'PCI Power Management Control and Status'
    }


class MID(Structure):
    _pack_ = 1
    _fields_ = [
        ('CID', c_uint16, 8),  # Capability ID (CID)
        ('NEXT', c_uint16, 8),  # Next Pointer (NEXT)
    ]
    desc = {'CID': 'Capability ID', 'NEXT': 'Next Pointer'}


class MC(Structure):
    _pack_ = 1
    _fields_ = [
        ('MSIE', c_uint16, 1),  # MSI Enable (MSIE)
        ('MMC', c_uint16, 3),  # Multiple Message Capable (MMC)
        ('MME', c_uint16, 3),  # Multiple Message Enable (MME)
        ('C64', c_uint16, 1),  # 64 Bit Address Capable (C64)
        ('PVM', c_uint16, 1),  # Per-Vector Masking Capable (PVM)
        ('RSV', c_uint16, 7),
    ]
    desc = {
        'MSIE': 'MSI Enable', 'MCC': 'Multiple Message Capable', 'MME': 'Multiple Message Enable',
        'C64': '64 Bit Address Capable', 'PVM': 'Per-Vector Masking Capable'
    }


class MA(Structure):
    _pack_ = 1
    _fields_ = [
        ('RSV', c_uint32, 2),
        ('ADDR', c_uint32, 30),  # Address (ADDR)
    ]
    desc = {'ADDR': 'Message Signaled Interrupt Message Address'}


class MSICAP(Structure):
    """System Bus (PCI Express) Registers - Message Signaled Interrupt Capability (Optional)"""
    _pack_ = 1
    _fields_ = [
        ('MID', MID),  # Message Signaled Interrupt Identifiers
        ('MC', MC),  # Message Signaled Interrupt Message Control
        ('MA', MA),  # Message Signaled Interrupt Message Address
        ('MUA', c_uint32),  # Message Signaled Interrupt Upper Address
        ('MD', c_uint16),  # Message Signaled Interrupt Message Data
        ('RSV', c_uint16),
        ('MMASK', c_uint32),  # Message Signaled Interrupt Mask Bits (Optional)
        ('MPEND', c_uint32),  # Message Signaled Interrupt Pending Bits (Optional)
    ]
    desc = {
        'MUA': 'Message Signaled Interrupt Upper Address', 'MD': 'Message Signaled Interrupt Message Data',
        'MMASK': 'Message Signaled Interrupt Mask Bits', 'MPEND': 'Message Signaled Interrupt Pending Bits'
    }


class MIDX(Structure):
    _pack_ = 1
    _fields_ = [
        ('CID', c_uint16, 8),  # Capability ID (CID)
        ('NEXT', c_uint16, 8),  # Next Pointer (NEXT)
    ]
    desc = {'CID': 'Capability ID', 'NEXT': 'Next Pointer'}


class MXC(Structure):
    _pack_ = 1
    _fields_ = [
        ('TS', c_uint16, 11),  # Table Size (TS)
        ('RSV', c_uint16, 3),
        ('FM', c_uint16, 1),  # Function Mask (FM)
        ('MXE', c_uint16, 1),  # MSI-X Enable (MXE)
    ]
    desc = {'TS': 'Table Size', 'FM': 'Function Mask', 'MXE': 'MSI-X Enable'}


class MTAB(Structure):
    _pack_ = 1
    _fields_ = [
        ('TBIR', c_uint32, 3),  # Table BIR (TBIR)
        ('TO', c_uint32, 29),  # Table Offset (TO)
    ]
    desc = {'TBIR': 'Table BIR', 'TO': 'Table Offset'}


class MPBA(Structure):
    _pack_ = 1
    _fields_ = [
        ('PBIR', c_uint32, 3),  # PBA Offset (PBAO)
        ('PBAO', c_uint32, 29),  # PBA BIR (PBIR)
    ]
    desc = {'TBIR': 'Table BIR', 'TO': 'Table Offset'}


class MSIXCAP(Structure):
    """System Bus (PCI Express) Registers - MSI-X Capability (Optional)"""
    _pack_ = 1
    _fields_ = [
        ('MIDX', MIDX),  # MSI-X Identifiers
        ('MXC', MXC),  # MSI-X Message Control
        ('MTAB', MTAB),  # MSI-X Table Offset / Table BIR
        ('MPBA', MPBA),  # MSI-X PBA Offset / PBA BIR
    ]
    desc = {
        'MIDX': 'MSI-X Identifiers', 'MXC': 'MSI-X Message Control',
        'MTAB': 'MSI-X Table Offset / Table BIR', 'MPBA': 'MSI-X PBA Offset / PBA BIR'
    }


class PXID(Structure):
    _pack_ = 1
    _fields_ = [
        ('CID', c_uint16, 3),  # Capability ID (CID)
        ('NEXT', c_uint16, 29),  # Next Pointer (NEXT)
    ]
    desc = {'CID': 'Capability ID', 'NEXT': 'Next Pointer'}


class PXCAP(Structure):
    _pack_ = 1
    _fields_ = [
        ('VER', c_uint16, 4),  # Capability Version (VER)
        ('DPT', c_uint16, 4),  # Device/Port Type (DPT)
        ('SI', c_uint16, 1),  # Slot Implemented (SI)
        ('IMN', c_uint16, 5),  # Interrupt Message Number (IMN)
        ('RSV', c_uint16, 2),
    ]
    desc = {
        'VER': 'Capability Version', 'DPT': 'Device/Port Type',
        'SI': 'Slot Implemented', 'IMN': 'Interrupt Message Number'
    }


class PXDCAP(Structure):
    _pack_ = 1
    _fields_ = [
        ('MPS', c_uint32, 3),  # Max_Payload_Size Supported (MPS)
        ('PFS', c_uint32, 2),  # Phantom Functions Supported (PFS)
        ('ETFS', c_uint32, 1),  # Extended Tag Field Supported (ETFS)
        ('L0SL', c_uint32, 3),  # Endpoint L0s Acceptable Latency (L0SL)
        ('L1L', c_uint32, 3),  # Endpoint L1 Acceptable Latency (L1L)
        ('RSV0', c_uint32, 3),
        ('RER', c_uint32, 1),  # Role-based Error Reporting (RER)
        ('RSV1', c_uint32, 2),
        ('CSPLV', c_uint32, 8),  # Captured Slot Power Limit Value (CSPLV)
        ('CSPLS', c_uint32, 2),  # Captured Slot Power Limit Scale (CSPLS)
        ('FLRC', c_uint32, 1),  # Function Level Reset Capability (FLRC)
        ('RSV2', c_uint32, 3),
    ]
    desc = {
        'MPS': 'Max_Payload_Size Supported', 'PFS': 'Phantom Functions Supported',
        'ETFS': 'Extended Tag Field Supported', 'L0SL': 'Endpoint L0s Acceptable Latency',
        'L1L': 'Endpoint L1 Acceptable Latency', 'RER': 'Role-based Error Reporting',
        'CSPLV': 'Captured Slot Power Limit Value', 'CSPLS': 'Captured Slot Power Limit Scale',
        'FLRC': 'Function Level Reset Capability'
    }


class PXDC(Structure):
    _pack_ = 1
    _fields_ = [
        ('CERE', c_uint16, 1),  # Correctable Error Reporting Enable (CERE)
        ('NFERE', c_uint16, 1),  # Non-Fatal Error Reporting Enable (NFERE)
        ('FERE', c_uint16, 1),  # Fatal Error Reporting Enable (FERE)
        ('URRE', c_uint16, 1),  # Unsupported Request Reporting Enable (URRE)
        ('ERO', c_uint16, 1),  # Enable Relaxed Ordering (ERO)
        ('MPS', c_uint16, 3),  # Max_Payload_Size (MPS)
        ('ETE', c_uint16, 1),  # Extended Tag Enable (ETE)
        ('PFE', c_uint16, 1),  # Phantom Functions Enable (PFE)
        ('APPME', c_uint16, 1),  # AUX Power PM Enable (APPME)
        ('ENS', c_uint16, 1),  # Enable No Snoop (ENS)
        ('MRRS', c_uint16, 3),  # Max_Read_Request_Size (MRRS)
        ('IFLR', c_uint16, 1),  # Initiate Function Level Reset
    ]
    desc = {
        'CERE': 'Correctable Error Reporting Enable', 'NFERE': 'Non-Fatal Error Reporting Enable',
        'FERE': 'Fatal Error Reporting Enable', 'URRE': 'Unsupported Request Reporting Enable',
        'ERO': 'Enable Relaxed Ordering', 'MPS': 'Max Payload Size', 'ETE': 'Extended Tag Enable',
        'PFE': 'Phantom Functions Enable', 'APPME': 'AUX Power PM Enable', 'ENS': 'Enable No Snoop',
        'MRRS': 'Max_Read_Request_Size', 'IFLR': 'Initiate Function Level Reset',
    }


class PXDS(Structure):
    _pack_ = 1
    _fields_ = [
        ('CED', c_uint16, 1),  # Correctable Error Detected (CED)
        ('NFED', c_uint16, 1),  # Non-Fatal Error Detected (NFED)
        ('FED', c_uint16, 1),  # Fatal Error Detected (FED)
        ('URD', c_uint16, 1),  # Unsupported Request Detected (URD)
        ('APD', c_uint16, 1),  # AUX Power Detected (APD)
        ('TP', c_uint16, 1),  # Transactions Pending (TP)
        ('RSV', c_uint16, 10),
    ]
    desc = {
        'CED': 'Correctable Error Detected', 'NFED': 'Non-Fatal Error Detected', 'FED': 'Fatal Error Detected',
        'URD': 'Unsupported Request Detected', 'APD': 'AUX Power Detected', 'TP': 'Transactions Pending'
    }


class PXLCAP(Structure):
    _pack_ = 1
    _fields_ = [
        ('SLS', c_uint32, 4),  # Supported Link Speeds (SLS)
        ('MLW', c_uint32, 6),  # Maximum Link Width (MLW)
        ('ASPMS', c_uint32, 2),  # Active State Power Management Support (ASPMS)
        ('L0SEL', c_uint32, 3),  # L0s Exit Latency (L0SEL)
        ('L1EL', c_uint32, 3),  # L1 Exit Latency (L1EL)
        ('CPM', c_uint32, 1),  # Clock Power Management (CPM)
        ('SDERC', c_uint32, 1),  # Surprise Down Error Reporting Capable (SDERC)
        ('DLLLA', c_uint32, 1),  # Data Link Layer Link Active Reporting Capable (DLLLA)
        ('LBNC', c_uint32, 1),  # Link Bandwidth Notification Capability (LBNC)
        ('AOC', c_uint32, 1),  # ASPM Optionality Compliance (AOC)
        ('RSV', c_uint32, 1),  #
        ('PN', c_uint32, 8),  # Port Number (PN)
    ]
    desc = {
        'SLS': 'Supported Link Speeds', 'MLW': 'Maximum Link Width', 'ASPMS': 'Active State Power Management Support',
        'L0SEL': 'L0s Exit Latency', 'L1EL': 'L1 Exit Latency', 'CPM': 'Clock Power Management',
        'SDERC': 'Surprise Down Error Reporting Capable', 'DLLLA': 'Data Link Layer Link Active Reporting Capable',
        'LBNC': 'Link Bandwidth Notification Capability', 'AOC': 'ASPM Optionality Compliance', 'PN': 'Port Number'
    }


class PXLC(Structure):
    _pack_ = 1
    _fields_ = [
        ('ASPMC', c_uint16, 2),  # Active State Power Management Control (ASPMC)
        ('RSV0', c_uint16, 1),
        ('RCB', c_uint16, 1),  # Read Completion Boundary (RCB)
        ('RSV1', c_uint16, 2),
        ('CCC)', c_uint16, 1),  # Common Clock Configuration (CCC)
        ('ES', c_uint16, 1),  # Extended Synch (ES)
        ('ECPM', c_uint16, 1),  # Enable Clock Power Management (ECPM)
        ('HAWD', c_uint16, 1),  # Hardware Autonomous Width Disable (HAWD)
        ('RSV2', c_uint16, 6),
    ]
    desc = {
        'ASPMC': 'Active State Power Management Control', 'RCB': 'Read Completion Boundary',
        'CCC': 'Common Clock Configuration', 'ES': 'Extended Synch',
        'ECPM': 'Enable Clock Power Management', 'HAWD': 'Hardware Autonomous Width Disable',
    }


class PXLS(Structure):
    _pack_ = 1
    _fields_ = [
        ('CLS', c_uint16, 4),  # Current Link Speed
        ('NLW', c_uint16, 6),  # Negotiated Link Width
        ('RSV1', c_uint16, 2),  # Reserved
        ('SCC', c_uint16, 1),  # Slot Clock Configuration
        ('RSV0', c_uint16, 3),  # Reserved
    ]
    desc = {
        'CLS': 'Current Link Speed ', 'NLW': 'Negotiated Link Width ', 'SCC': 'Slot Clock Configuration'
    }


class PXDCAP2(Structure):
    _pack_ = 1
    _fields_ = [
        ('CTRS', c_uint32, 4),  # Completion Timeout Ranges Supported
        ('CTDS', c_uint32, 1),  # Completion Timeout Disable Supported
        ('ARIFS', c_uint32, 1),  # ARI Forwarding Supported
        ('AORS', c_uint32, 1),  # AtomicOp Routing Supported
        ('32AOCS', c_uint32, 1),  # 32-bit AtomicOp Completer Supported
        ('64AOCS', c_uint32, 1),  # 64-bit AtomicOp Completer Supported
        ('128CCS', c_uint32, 1),  # 128-bit CAS Completer Supported
        ('NPRPR', c_uint32, 1),  # No RO-enabled PR-PR Passing
        ('LTRS', c_uint32, 1),  # Latency Tolerance Reporting Supported
        ('TPHCS', c_uint32, 2),  # TPH Completer Supported
        ('RSV1', c_uint32, 4),  # Reserved
        ('OBFFS', c_uint32, 2),  # OBFF Supported
        ('EFFS', c_uint32, 1),  # Extended Fmt Field Supported
        ('EETPS', c_uint32, 1),  # End-End TLP Prefix Supported
        ('MEETP', c_uint32, 2),  # Max End-End TLP Prefixes
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'CTRS': 'Completion Timeout Ranges Supported ', 'CTDS': 'Completion Timeout Disable Supported ',
        'ARIFS': 'ARI Forwarding Supported ', 'AORS': 'AtomicOp Routing Supported ',
        '32AOCS': '32-bit AtomicOp Completer Supported ', '64AOCS': '64-bit AtomicOp Completer Supported ',
        '128CCS': '128-bit CAS Completer Supported ', 'NPRPR': 'No RO-enabled PR-PR Passing ',
        'LTRS': 'Latency Tolerance Reporting Supported ', 'TPHCS': 'TPH Completer Supported ',
        'OBFFS': 'OBFF Supported ', 'EFFS': 'Extended Fmt Field Supported ', 'EETPS': 'End-End TLP Prefix Supported ',
        'MEETP': 'Max End-End TLP Prefixes '
    }


class PXDC2(Structure):
    _pack_ = 1
    _fields_ = [
        ('CTV', c_uint32, 4),  # Completion Timeout Value
        ('CTD', c_uint32, 1),  # Completion Timeout Disable
        ('RSV2', c_uint32, 5),  # Reserved
        ('LTRME', c_uint32, 1),  # Latency Tolerance Reporting Mechanism Enable
        ('RSV1', c_uint32, 2),  # Reserved
        ('OBFFE', c_uint32, 2),  # OBFF Enable
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'CTV': 'Completion Timeout Value', 'CTD': 'Completion Timeout Disable ',
        'LTRME': 'Latency Tolerance Reporting Mechanism Enable ', 'OBFFE': 'OBFF Enable '
    }


class PCIECAP(Structure):
    """System Bus (PCI Express) Registers - PCI Express Capability"""
    _pack_ = 1
    _fields_ = [
        ('PXID', PXID),  # PCI Express Capability ID
        ('PXCAP', PXCAP),  # PCI Express Capabilities
        ('PXDCAP', PXDCAP),  # PCI Express Device Capabilities
        ('PXDC', PXDC),  # PCI Express Device Control
        ('PXDS', PXDS),  # PCI Express Device Status
        ('PXLCAP', PXLCAP),  # PCI Express Link Capabilities
        ('PXLC', PXLC),  # PCI Express Link Control
        ('PXLS', PXLS),  # PCI Express Link Status
        ('PXDCAP2', PXDCAP2),  # PCI Express Device Capabilities 2
        ('PXDC2', PXDC2),  # PCI Express Device Control 2
    ]
    desc = {
        'PXID': 'PCI Express Capability ID', 'PXCAP': 'PCI Express Capabilities',
        'PXDCAP': 'PCI Express Device Capabilities', 'PXDC': 'PCI Express Device Control',
        'PXDS': 'PCI Express Device Status', 'PXLCAP': 'PCI Express Link Capabilities',
        'PXLC': 'PCI Express Link Control', 'PXLS': 'PCI Express Link Status',
        'PXDCAP2': 'PCI Express Device Capabilities 2', 'PXDC2': 'PCI Express Device Control 2',
    }


class AERID(Structure):
    _pack_ = 1
    _fields_ = [
        ('CID', c_uint32, 16),  # Capability ID
        ('CVER', c_uint32, 4),  # Capability Version
        ('NEXT', c_uint32, 1),  # Next Pointer
    ]
    desc = {
        'CID': 'Capability ID ', 'CVER': 'Capability Version ', 'NEXT': 'Next Pointer '
    }


class AERUCES(Structure):
    _pack_ = 1
    _fields_ = [
        ('RSV2', c_uint32, 4),  # Reserved
        ('DLPES', c_uint32, 1),  # Data Link Protocol Error Status
        ('RSV1', c_uint32, 7),  # Reserved
        ('PTS', c_uint32, 1),  # Poisoned TLP Status
        ('FCPES', c_uint32, 1),  # Flow Control Protocol Error Status
        ('CTS', c_uint32, 1),  # Completion Timeout Status
        ('CAS', c_uint32, 1),  # Completer Abort Status
        ('UCS', c_uint32, 1),  # Unexpected Completion Status
        ('ROS', c_uint32, 1),  # Receiver Overflow Status
        ('MTS', c_uint32, 1),  # Malformed TLP Status
        ('ECRCES', c_uint32, 1),  # ECRC Error Status
        ('URES', c_uint32, 1),  # Unsupported Request Error Status
        ('ACSVS', c_uint32, 1),  # ACS Violation Status
        ('UIES', c_uint32, 1),  # Uncorrectable Internal Error Status
        ('MCBTS', c_uint32, 1),  # MC Blocked TLP Status
        ('AOEBS', c_uint32, 1),  # AtomicOp Egress Blocked Status
        ('TPBES', c_uint32, 1),  # TLP Prefix Blocked Error Status
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'DLPES': 'Data Link Protocol Error Status ', 'PTS': 'Poisoned TLP Status ',
        'FCPES': 'Flow Control Protocol Error Status ', 'CTS': 'Completion Timeout Status ',
        'CAS': 'Completer Abort Status ', 'UCS': 'Unexpected Completion Status ', 'ROS': 'Receiver Overflow Status ',
        'MTS': 'Malformed TLP Status ', 'ECRCES': 'ECRC Error Status ', 'URES': 'Unsupported Request Error Status ',
        'ACSVS': 'ACS Violation Status ', 'UIES': 'Uncorrectable Internal Error Status ',
        'MCBTS': 'MC Blocked TLP Status ', 'AOEBS': 'AtomicOp Egress Blocked Status ',
        'TPBES': 'TLP Prefix Blocked Error Status '
    }


class AERUCEM(Structure):
    _pack_ = 1
    _fields_ = [
        ('RSV2', c_uint32, 4),  # Reserved
        ('DLPEM', c_uint32, 1),  # Data Link Protocol Error Mask
        ('RSV1', c_uint32, 7),  # Reserved
        ('PTM', c_uint32, 1),  # Poisoned TLP Mask
        ('FCPEM', c_uint32, 1),  # Flow Control Protocol Error Mask
        ('CTM', c_uint32, 1),  # Completion Timeout Mask
        ('CAM', c_uint32, 1),  # Completer Abort Mask
        ('UCM', c_uint32, 1),  # Unexpected Completion Mask
        ('ROM', c_uint32, 1),  # Receiver Overflow Mask
        ('MTM', c_uint32, 1),  # Malformed TLP Mask
        ('ECRCEM', c_uint32, 1),  # ECRC Error Mask
        ('UREM', c_uint32, 1),  # Unsupported Request Error Mask
        ('ACSVM', c_uint32, 1),  # ACS Violation Mask
        ('UIEM', c_uint32, 1),  # Uncorrectable Internal Error Mask
        ('MCBTM', c_uint32, 1),  # MC Blocked TLP Mask
        ('AOEBM', c_uint32, 1),  # AtomicOp Egress Blocked Mask
        ('TPBEM', c_uint32, 1),  # TLP Prefix Blocked Error Mask
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'DLPEM': 'Data Link Protocol Error Mask ', 'PTM': 'Poisoned TLP Mask ',
        'FCPEM': 'Flow Control Protocol Error Mask ', 'CTM': 'Completion Timeout Mask ', 'CAM': 'Completer Abort Mask ',
        'UCM': 'Unexpected Completion Mask ', 'ROM': 'Receiver Overflow Mask ', 'MTM': 'Malformed TLP Mask ',
        'ECRCEM': 'ECRC Error Mask ', 'UREM': 'Unsupported Request Error Mask ', 'ACSVM': 'ACS Violation Mask ',
        'UIEM': 'Uncorrectable Internal Error Mask ', 'MCBTM': 'MC Blocked TLP Mask ',
        'AOEBM': 'AtomicOp Egress Blocked Mask ', 'TPBEM': 'TLP Prefix Blocked Error Mask '
    }


class AERUCESEV(Structure):
    _pack_ = 1
    _fields_ = [
        ('RSV2', c_uint32, 4),  # Reserved
        ('DLPESEV', c_uint32, 1),  # Data Link Protocol Error Severity
        ('RSV1', c_uint32, 7),  # Reserved
        ('PTSEV', c_uint32, 1),  # Poisoned TLP Severity
        ('FCPESEV', c_uint32, 1),  # Flow Control Protocol Error Severity
        ('CTSEV', c_uint32, 1),  # Completion Timeout Severity
        ('CASEV', c_uint32, 1),  # Completer Abort Severity
        ('UCSEV', c_uint32, 1),  # Unexpected Completion Severity
        ('ROSEV', c_uint32, 1),  # Receiver Overflow Severity
        ('MTSEV', c_uint32, 1),  # Malformed TLP Severity
        ('ECRCESEV', c_uint32, 1),  # ECRC Error Severity
        ('URESEV', c_uint32, 1),  # Unsupported Request Error Severity
        ('ACSVSEV', c_uint32, 1),  # ACS Violation Severity
        ('UIESEV', c_uint32, 1),  # Uncorrectable Internal Error Severity
        ('MCBTSEV', c_uint32, 1),  # MC Blocked TLP Severity
        ('AOEBSEV', c_uint32, 1),  # AtomicOp Egress Blocked Severity
        ('TPBESEV', c_uint32, 1),  # TLP Prefix Blocked Error Severity
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'DLPESEV': 'Data Link Protocol Error Severity ', 'PTSEV': 'Poisoned TLP Severity ',
        'FCPESEV': 'Flow Control Protocol Error Severity ', 'CTSEV': 'Completion Timeout Severity ',
        'CASEV': 'Completer Abort Severity ', 'UCSEV': 'Unexpected Completion Severity ',
        'ROSEV': 'Receiver Overflow Severity ', 'MTSEV': 'Malformed TLP Severity ', 'ECRCESEV': 'ECRC Error Severity ',
        'URESEV': 'Unsupported Request Error Severity ', 'ACSVSEV': 'ACS Violation Severity ',
        'UIESEV': 'Uncorrectable Internal Error Severity ', 'MCBTSEV': 'MC Blocked TLP Severity ',
        'AOEBSEV': 'AtomicOp Egress Blocked Severity ', 'TPBESEV': 'TLP Prefix Blocked Error Severity '
    }


class AERCES(Structure):
    _pack_ = 1
    _fields_ = [
        ('RES', c_uint32, 1),  # Receiver Error Status
        ('RSV2', c_uint32, 5),  # Reserved
        ('BTS', c_uint32, 1),  # Bad TLP Status
        ('BDS', c_uint32, 1),  # Bad DLLP Status
        ('RRS', c_uint32, 1),  # REPLAY
        ('RSV1', c_uint32, 3),  # Reserved
        ('RTS', c_uint32, 1),  # Replay Timer Timeout Status
        ('ANFES', c_uint32, 1),  # Advisory Non-Fatal Error Status
        ('CIES', c_uint32, 1),  # Corrected Internal Error Status
        ('HLOS', c_uint32, 1),  # Header Log Overflow Status
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'RES': 'Receiver Error Status ', 'BTS': 'Bad TLP Status ', 'BDS': 'Bad DLLP Status ', 'RRS': 'REPLAY',
        'RTS': 'Replay Timer Timeout Status ', 'ANFES': 'Advisory Non-Fatal Error Status ',
        'CIES': 'Corrected Internal Error Status ', 'HLOS': 'Header Log Overflow Status '
    }


class AERCEM(Structure):
    _pack_ = 1
    _fields_ = [
        ('REM', c_uint32, 1),  # Receiver Error Mask
        ('RSV2', c_uint32, 5),  # Reserved
        ('BTM', c_uint32, 1),  # Bad TLP Mask
        ('BDM', c_uint32, 1),  # Bad DLLP Mask
        ('RRM', c_uint32, 1),  # REPLAY
        ('RSV1', c_uint32, 3),  # Reserved
        ('RTM', c_uint32, 1),  # Replay Timer Timeout Mask
        ('ANFEM', c_uint32, 1),  # Advisory Non-Fatal Error Mask
        ('CIEM', c_uint32, 1),  # Corrected Internal Error Mask
        ('HLOM', c_uint32, 1),  # Header Log Overflow Mask
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'REM': 'Receiver Error Mask ', 'BTM': 'Bad TLP Mask ', 'BDM': 'Bad DLLP Mask ', 'RRM': 'REPLAY',
        'RTM': 'Replay Timer Timeout Mask ', 'ANFEM': 'Advisory Non-Fatal Error Mask ',
        'CIEM': 'Corrected Internal Error Mask ', 'HLOM': 'Header Log Overflow Mask '
    }


class AERCC(Structure):
    _pack_ = 1
    _fields_ = [
        ('FEP', c_uint32, 5),  # First Error Pointer
        ('EGC', c_uint32, 1),  # ECRC Generation Capable
        ('EGE', c_uint32, 1),  # ECRC Generation Enable
        ('ECC', c_uint32, 1),  # ECRC Check Capable
        ('ECE', c_uint32, 1),  # ECRC Check Enable
        ('MHRC', c_uint32, 1),  # Multiple Header Recording Capable
        ('MHRE', c_uint32, 1),  # Multiple Header Recording Enable
        ('TPLP', c_uint32, 1),  # TLP Prefix Log Present
        ('RSV0', c_uint32, 1),  # Reserved
    ]
    desc = {
        'FEP': 'First Error Pointer ', 'EGC': 'ECRC Generation Capable ', 'EGE': 'ECRC Generation Enable ',
        'ECC': 'ECRC Check Capable ', 'ECE': 'ECRC Check Enable ', 'MHRC': 'Multiple Header Recording Capable ',
        'MHRE': 'Multiple Header Recording Enable ', 'TPLP': 'TLP Prefix Log Present '
    }


class AERHL(Structure):
    _pack_ = 1
    _fields_ = [
        ('HB3', c_uint8),  # Header Byte 3
        ('HB2', c_uint8),  # Header Byte 2
        ('HB1', c_uint8),  # Header Byte 1
        ('HB0', c_uint8),  # Header Byte 0
        ('HB7', c_uint8),  # Header Byte 7
        ('HB6', c_uint8),  # Header Byte 6
        ('HB5', c_uint8),  # Header Byte 5
        ('HB4', c_uint8),  # Header Byte 4
        ('HB11', c_uint8),  # Header Byte 11
        ('HB10', c_uint8),  # Header Byte 10
        ('HB9', c_uint8),  # Header Byte 9
        ('HB8', c_uint8),  # Header Byte 8
        ('HB15', c_uint8),  # Header Byte 15
        ('HB14', c_uint8),  # Header Byte 14
        ('HB13', c_uint8),  # Header Byte 13
        ('HB12', c_uint8),  # Header Byte 12
    ]
    desc = {
        'HB3': 'Header Byte 3 ', 'HB2': 'Header Byte 2 ', 'HB1': 'Header Byte 1 ', 'HB0': 'Header Byte 0 ',
        'HB7': 'Header Byte 7 ', 'HB6': 'Header Byte 6 ', 'HB5': 'Header Byte 5 ', 'HB4': 'Header Byte 4 ',
        'HB11': 'Header Byte 11 ', 'HB10': 'Header Byte 10 ', 'HB9': 'Header Byte 9 ', 'HB8': 'Header Byte 8 ',
        'HB15': 'Header Byte 15 ', 'HB14': 'Header Byte 14 ', 'HB13': 'Header Byte 13 ', 'HB12': 'Header Byte 12 '
    }


class AERTLP(Structure):
    _pack_ = 1
    _fields_ = [
        ('TPL1B3', c_uint8),  # First TLP Prefix Log Byte 3
        ('TPL1B2', c_uint8),  # First TLP Prefix Log Byte 2
        ('TPL1B1', c_uint8),  # First TLP Prefix Log Byte 1
        ('TPL1B0', c_uint8),  # First TLP Prefix Log Byte 0
        ('TPL2B3', c_uint8),  # Second TLP Prefix Log Byte 3
        ('TPL2B2', c_uint8),  # Second TLP Prefix Log Byte 2
        ('TPL2B1', c_uint8),  # Second TLP Prefix Log Byte 1
        ('TPL2B0', c_uint8),  # Second TLP Prefix Log Byte 0
        ('TPL3B3', c_uint8),  # Third TLP Prefix Log Byte 3
        ('TPL3B2', c_uint8),  # Third TLP Prefix Log Byte 2
        ('TPL3B1', c_uint8),  # Third TLP Prefix Log Byte 1
        ('TPL3B0', c_uint8),  # Third TLP Prefix Log Byte 0
        ('TPL4B3', c_uint8),  # Fourth TLP Prefix Log Byte 3
        ('TPL4B2', c_uint8),  # Fourth TLP Prefix Log Byte 2
        ('TPL4B1', c_uint8),  # Fourth TLP Prefix Log Byte 1
        ('TPL4B0', c_uint8),  # Fourth TLP Prefix Log Byte 0
    ]
    desc = {
        'TPL1B3': 'First TLP Prefix Log Byte 3 ', 'TPL1B2': 'First TLP Prefix Log Byte 2 ',
        'TPL1B1': 'First TLP Prefix Log Byte 1 ', 'TPL1B0': 'First TLP Prefix Log Byte 0 ',
        'TPL2B3': 'Second TLP Prefix Log Byte 3 ', 'TPL2B2': 'Second TLP Prefix Log Byte 2 ',
        'TPL2B1': 'Second TLP Prefix Log Byte 1 ', 'TPL2B0': 'Second TLP Prefix Log Byte 0 ',
        'TPL3B3': 'Third TLP Prefix Log Byte 3 ', 'TPL3B2': 'Third TLP Prefix Log Byte 2 ',
        'TPL3B1': 'Third TLP Prefix Log Byte 1 ', 'TPL3B0': 'Third TLP Prefix Log Byte 0 ',
        'TPL4B3': 'Fourth TLP Prefix Log Byte 3 ', 'TPL4B2': 'Fourth TLP Prefix Log Byte 2 ',
        'TPL4B1': 'Fourth TLP Prefix Log Byte 1 ', 'TPL4B0': 'Fourth TLP Prefix Log Byte 0 '
    }


class AERCAP(Structure):
    """System Bus (PCI Express) Registers - Advanced Error Reporting Capability (Optional)"""
    _pack_ = 1
    _fields_ = [
        ('AERID', AERID),  # AER Capability ID
        ('AERUCES', AERUCES),  # AER Uncorrectable Error Status Register
        ('AERUCEM', AERUCEM),  # AER Uncorrectable Error Mask Register
        ('AERUCESEV', AERUCESEV),  # AER Uncorrectable Error Severity Register
        ('AERCES', AERCES),  # AER Correctable Error Status Register
        ('AERCEM', AERCEM),  # AER Correctable Error Mask Register
        ('AERCC', AERCC),  # AER Advanced Error Capabilities and Control Reg
        ('AERHL', AERHL),  # AER Header Log Register
        ('AERTLP', AERTLP),  # AER TLP Prefix Log Register (Optional)
    ]
    desc = {
        'AERID': 'AER Capability ID', 'AERUCES': 'AER Uncorrectable Error Status Register',
        'AERUCEM': 'AER Uncorrectable Error Mask Register', 'AERUCESEV': 'AER Uncorrectable Error Severity Register',
        'AERCES': 'AER Correctable Error Status Register', 'AERCEM': 'AER Correctable Error Mask Register',
        'AERCC': 'AER Advanced Error Capabilities and Control Reg', 'AERHL': 'AER Header Log Register',
        'AERTLP': 'AER TLP Prefix Log Register'
    }
